-- Cross Platform Music Compiler
-- /Mic, 2008

-- This is the compiler part of XPMCK, which apart from the compiler also contains
-- playback libraries for the supported targets.

without warning

include file.e
include get.e
include wildcard.e
include machine.e

constant	TARGET_UNKNOWN = 0,
		TARGET_SMS = 1,
		TARGET_NES = 2,
		TARGET_GBA = 3,
		TARGET_NDS = 4,
		TARGET_GBC = 5,
		TARGET_MGD = 6,
		TARGET_SAT = 7,
		TARGET_XGS = 8,
		TARGET_SGG = 9

constant	CMD_NOTE   = #00,
                CMD_REST   = #0C,
                CMD_OCTAVE = #10,
		CMD_DUTY   = #20,
		CMD_VOL2   = #30,
		CMD_OCTUP  = #40,
		CMD_OCTDN  = #50,
		CMD_VOLOUP = #60,
		CMD_VOLODN = #70,
		CMD_ARP2   = #80,
		CMD_ARPOFF = #90,
		CMD_CBOFF  = #E0,
		CMD_CBONCE = #E1,
		CMD_CBEVNT = #E2,
		CMD_CBEVVC = #E3,
		CMD_CBEVVM = #E4,
		CMD_CBEVOC = #E5,
		CMD_CBNOTE = #E6,
		CMD_HWTE   = #E8,
		CMD_HWVE   = #E9,
		CMD_LDWAVE = #EC,
		CMD_DETUNE = #ED,
		CMD_VOLSET = #F0,
		CMD_VOLMAC = #F1,
		CMD_DUTMAC = #F2,
		CMD_PORMAC = #F3,
		CMD_PANMAC = #F4,
		CMD_VIBMAC = #F5,
		CMD_SWPMAC = #F6,
		CMD_VSLMAC = #F7,
		CMD_ARPMAC = #F8,
		CMD_JMP    = #F9,
		CMD_DJNZ   = #FA,
		CMD_LOPCNT = #FB,
		CMD_APMAC2 = #FC,
		CMD_END	   = #FF
		
		
constant	NOTES = {'c', -2, 'd', -2, 'e', 'f', -2, 'g', -2, 'a', -2, 'b', 'r'}

-- Octave 1 frequencies
constant 	OCTAVE1 =
{
	32.7032,	-- C
	34.6479,	-- C+ / D-
	36.7081,	-- D
	38.8909,	-- D+ / E-
	41.2035,	-- E
	43.6536,	-- F
	46.2493,	-- F+ / G-
	48.9995,	-- G
	51.9130,	-- G+ / A-
	55.0000,	-- A
	58.2705,	-- A+ / B-
	61.7354		-- B
}

-- Alternative frequencies for the SEGA Master System PSG
constant 	OCTAVE1_ALT_SMS = OCTAVE1 * 0.99370454545454545454545454545455


constant	SUPPORTED_LENGTHS = {32, 16, 8, 4, 2, 1}


constant	DIGITS = "0123456789",
		ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
		ALPHANUM = DIGITS & ALPHABET & lower(ALPHABET),
		ALPHANUM2 = ALPHANUM & '@'

constant	MONTHS = {"January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"}
constant	WEEKDAYS = {"Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"}


atom		updateFreq

integer 	target,
		verbose,
		warningsAreErrors,
		fileDataPos,
		inFile,
		outFile,
		lineNum,
		songNum,
		maxTempo,
		minNote,
		minOctave,
		maxOctave,
		minVolume,
		maxWavLength,
		maxWavSample,
		minWavSample,
		supportsPan,		-- Target supports some form of panning (@CS)
		supportsPAL,		-- Target has a 50 Hz video mode
		octaveRev,
		enRev,
		rest,
		maxLoopDepth,
		tune,
		useFractionalDelays	-- Use fixed-point note lengths with fractional part (allows for
					-- finer granularity is tempo settings).
		

sequence 	cmdLine,
		fileData,
		restOfLine,
		shortFilename,
		supportedChannels,	-- Channel names supported on the target
		songTitle,
		songComposer,
		songProgrammer,
		activeChannels,		-- Currently active channels
		supportsDutyChange,	-- Channel supports variable duty cycles
		supportsVolumeChange,	-- Channel volume can be changed (not just be set too on/off)
		supportsPCM,		-- Fixed frequency, samples tied to specific notes
		supportsWave,		-- Variable frequency
		supportsHwToneEnv,
		supportsHwVolEnv,
		supportsDetune,
		maxVolume,
		machineVolLimit,
		macros,
		loopStack,
		volumeMacros,
		panMacros,
		dutyMacros,
		pitchMacros,
		arpeggios,
		vibratos,
		portamentos,
		volumeSlides,
		callbacks,
		waveforms,
		tuple,
		pcms,
		usesEN,
		currentLength,
		currentOctave,
		currentTempo,
		currentCutoff,
		currentNote,
		currentVolume,
		loopPoint,
		lastSetLength,
		pendingOctChange,
		songs,
		songLen,
		songLoopLen,
		hasAnyNote,
		defines
		


-- Output an error message and abort execution
procedure ERROR(sequence msg, integer line)
	if line = -1 then
		puts(1, "Error: " & msg & "\n")
	else
		printf(1, "%d, Error: " & msg & "\n", line)
	end if
	abort(0)
end procedure


-- Output a warning message
procedure WARNING(sequence msg, integer line)
	if warningsAreErrors then
		ERROR(msg, line)
	else
		if line = -1 then
			puts(1, "Warning: " & msg & "\n")
		else
			printf(1, "%d, Warning: " & msg & "\n", line)
		end if
	end if
end procedure


function abs(atom a)
	if a < 0 then
		return -a
	end if
	return a
end function


function getch()
	integer c
	
	c = -1
	if fileDataPos <= length(fileData) then
		c = fileData[fileDataPos]
		fileDataPos += 1
	end if
	
	return c
end function


-- Unget the last read character  
procedure ungetch()
	if fileDataPos > 1 then
		fileDataPos -= 1
	end if
end procedure


-- Consume whitespace
procedure skip_whitespace()
	integer c
	
	c = 0
	while c != -1 do
		c = getch()
		if c = ' ' or c = '\t' or c = 13 or c = 10 then
			if c = 10 then
				lineNum += 1
			end if
		else
			exit
		end if
	end while
	
	ungetch()
end procedure


-- Sum up all elements of s
function sum(sequence s)
	integer n
	
	n = 0
	for i = 1 to length(s) do
		n += s[i]
	end for
	
	return n
end function


-- Push o onto s and return s
function push(sequence s, object o)
	return append(s, o)
end function


-- Pop the top element from s and return {top of s, rest of s}
function pop(sequence s)
	object o
	
	o = 0
	
	if length(s) then
		o = s[length(s)]
		if length(s) > 1 then
			s = s[1..length(s) - 1]
		else
			s = {}
		end if
	end if
	
	return {o, s}
end function


-- Check if o is in the range of min and max
--
-- Examples:
--   in_range(1, 0, 5) -> 1
--   in_range({1,2,3}, 1, 10) -> 1
--   in_range({1,2}, {0,0}, 5) -> 1
--   in_range({1,2}, {2,0}, 5) -> 0     (min[1]>o[1])
--   in range({1,2}, {0,0,0}, 5) -> 0   (too many elements in min)
--
function in_range(object o, object min, object max)
	if atom(o) then
		if atom(min) and atom(max) then
			return (o >= min and o <= max)
		else
			return 0
		end if
	else
		if atom(min) and atom(max) then
			for i = 1 to length(o) do
				if o[i] < min or o[i] > max then
					return 0
				end if
			end for
		elsif atom(min) and sequence(max) then
			if length(max) = length(o) then
				for i = 1 to length(o) do
					if o[i] < min or o[i] > max[i] then
						return 0
					end if
				end for
			else
				return 0
			end if
		elsif sequence(min) and atom(max) then
			if length(min) = length(o) then
				for i = 1 to length(o) do
					if o[i] < min[i] or o[i] > max then
						return 0
					end if
				end for
			else
				return 0
			end if
		else
			if length(min) = length(max) and length(max) = length(o) then
				for i = 1 to length(o) do
					if o[i] < min[i] or o[i] > max[i] then
						return 0
					end if
				end for
			else
				return 0
			end if
		end if
	end if
	
	return 1
end function


-- Return true (1) if o is a digit or a sequence of digits
function is_numeric(object o)
	return in_range(o, '0', '9')
end function


-- Return true (1) if o is a letter in the english alphabet or a sequence of such letters
function is_alpha(object o)
	return in_range(o, 'a', 'z') or in_range(o, 'A', 'Z')
end function


function is_alphanum(object o)
	if atom(o) then
		return (is_alpha(o) or is_numeric(o))
	end if
	for i = 1 to length(o) do
		if ((not is_alpha(o[i])) and (not is_numeric(o[i]))) then
			return 0
		end if
	end for
	return 1
end function


-- Read a string (anything but whitespace or EOF)
function get_string()
	sequence s
	integer c
	
	skip_whitespace()
	
	s = ""
	c = 0
	while c != -1 do
		c = getch()
		if c = -1 or c = ' ' or c = '\t' or c = 13 or c = 10 then
			exit
		else
			s &= c
		end if
	end while
	
	ungetch()
	
	return s
end function


-- Read a string of characters that belong to the set specified in validChars
function get_string_in_range(sequence validChars)
	sequence s
	integer c
	
	skip_whitespace()

	s = ""
	c = 0
	while c != -1 do
		c = getch()
		if find(c, validChars) <= 0 then
			exit
		else
			s &= c
		end if
	end while
	
	ungetch()
	
	return s
end function


-- Read a string of characters until some character in endChars is found
function get_string_until(sequence endChars)
	sequence s
	integer c
	
	skip_whitespace()
	s = ""
	c = 0
	while c != -1 do
		c = getch()
		if find(c, endChars) > 0 then
			exit
		else
			s &= c
		end if
	end while
	
	ungetch()
	
	return s
end function


-- Read a string of digits
function get_numeric_string()
	sequence s
	integer c
	
	skip_whitespace()

	s = ""
	c = 0
	while c != -1 do
		c = getch()
		if (c >= '0' and c <= '9') or 
		   (c = '-' and not length(s)) then
			s &= c
		else
			exit
		end if
	end while
	
	ungetch()
	
	return s
end function


-- Read a string of alphanumeric characters
function get_alphanum_string()
	sequence s
	integer c
	
	skip_whitespace()

	s = ""
	c = 0
	while c != -1 do
		c = getch()
		if (c >= '0' and c <= '9') or
		   (c >= 'A' and c <= 'Z') or
		   (c >= 'a' and c <= 'z') then
			s &= c
		else
			exit
		end if
	end while
	
	ungetch()
	
	return s
end function
	

-- Return the substring of s from first..last
function substr(sequence s, integer first, integer last)
	sequence t
	
	t = ""
	
	-- Make sure the indices are ok
	if first < 1 then
		first = 1
	end if
	if first > length(s) then
		first = length(s)
	end if
	if last < 1 then
		last = 1
	end if
	if last > length(s) then
		last = length(s)
	end if
	
	-- The slice must have positive length
	if first <= last then
		t = s[first..last]
	end if
	
	return t
end function


-- Split an alphanumeric string into {alpha_part, numeric_part}
function split_alphanum_string(sequence s)
	integer j
	
	j = 0
	for i = 1 to length(s) do
		if s[i] >= '0' and s[i] <= '9' then
			j = i
			exit
		end if
	end for
	
	return {substr(s, 1, j-1), substr(s, j, length(s))}
end function
	

-- Define a symbol with a given name and value
procedure define(sequence what, object val)
	integer i
	
	i = find(what, defines[1])
	if i >= 1 then
		defines[2][i] = val
	else
		defines[1] = append(defines[1], what)
		defines[2] = append(defines[2], val)
	end if
end procedure


-- Check if a symbol with a given name is defined
function is_defined(sequence what)
	return find(what, defines[1]) >= 1
end function


-- Read a {}-enclosed list of numbers separated by spaces or commas. The list can
-- contain a loop, beginning at | and ranging to the end of the list.
function get_list()
	sequence s,t
	integer c, commaOk, endOk, pipeOk, concatTo
	object o
	
	-- s[2] will contain everything before the |, s[3] everything after it.
	s = {GET_FAIL, {}, {}}
	
	commaOk = 0	-- Not ok to read a comma
	pipeOk = 1	-- Ok to read a |
	endOk = 0	-- Not ok to read a }
	concatTo = 2	-- Concatenate to s[2]
	
	skip_whitespace()
	
	c = getch()
	
	if c = '{' then
		while 1 do
			skip_whitespace()
			t = get_string_in_range(DIGITS & '-')
			if length(t) then
				o = value(t)
				if o[1] = GET_SUCCESS then
					s[concatTo] &= o[2]
					commaOk = 1
					pipeOk = 1
					endOk = 1
				else
					ERROR("Syntax error: " & t, lineNum)
				end if
			else
				c = getch()
				if c = ',' then
					if not commaOk then
						ERROR("Unexpected comma", lineNum)
					end if
					commaOk = 0
					pipeOk = 0
					endOk = 0
				elsif c = '|' then
					if concatTo = 2 and pipeOk then
						concatTo = 3
						commaOk = 0
						pipeOk = 0
						endOk = 0
					else
						ERROR("Unexpected |", lineNum)
					end if
				elsif c = '}' then
					if endOk then
						s[1] = GET_SUCCESS
						exit
					else
						ERROR("Malformed list", lineNum)
					end if
				end if
			end if
		end while
	else
		ERROR("Expected {, got " & c, lineNum)
	end if
			
	return s
end function


-- Step through a list
-- Return {new_position next_item}
function step_list(sequence l, sequence pos)
	sequence t
	-- ? l[pos[2]]
	if pos[1] <= length(l[pos[2]]) then
		t = {l[pos[2]][pos[1]]}
		pos[1] += 1
		t = pos[1..2] & t
	else --pos[2] < 3 then
		if length(l[3]) then
			pos[2] = 3
			pos[1] = 1
			t = {l[pos[2]][pos[1]]}
			pos[1] += 1
			t = pos[1..2] & t
		else
			t = {length(l[2]), 2, l[2][length(l[2])]}
		end if
	--else
	--	t = {}
	end if
	
	return t
end function
		
	
-- Return a sequence containing a text-form representation on a list returned by get_list
-- 
-- Examples:
--   {GET_SUCCESS, {1, 2, 3}, {}} -> "{1 2 3}"
--   {GET_SUCCESS, {1, 2, 3}, {4, 5}} -> "{1 2 3 | 4 5}"
--   {GET_SUCCESS, {}, {4, 5}} -> "{| 4 5}"
--
function sprint_list(sequence l)
	sequence t
	
	t = "{"
	for i = 1 to length(l[2]) do
		t &= sprintf("%d", l[2][i])
		if i < length(l[2]) then
			t &= ' '
		end if
	end for
	
	if length(l[3]) then
		t &= " | "
		for i = 1 to length(l[3]) do
			t &= sprintf("%d", l[3][i])
			if i < length(l[3]) then
				t &= ' '
			end if
		end for
	end if
	
	return t & '}'
end function


function is_empty_list(sequence l)
	return (length(l[2]) + length(l[3])) = 0
end function

	

-- Output all pending notes to the current song sequence			
procedure write_note(integer forceOctChange)
	atom frames, cutoffFrames, scaling
	integer length32, length256, w1, w2
	
	for i = 1 to length(supportedChannels) do
		if length(currentNote[i]) then
			if integer(tuple[i]) then
				songLen[songNum][i] += currentNote[i][2]
				frames = (updateFreq * 60.0) / currentTempo[i]			-- frames per quarternote

				if useFractionalDelays then
					length32 = floor((frames / 8) * 256)
					scaling = 256
					--frames *= currentNote[i][2] / 8
					frames = length32 * currentNote[i][2]
					if currentCutoff[i] < 0 then
						cutoffFrames = -currentCutoff[i] * scaling
						if cutoffFrames > frames then
							cutoffFrames = frames
						end if
					else
						cutoffFrames = (frames * (8 - currentCutoff[i])) / 8
					end if
					frames -= cutoffFrames
					if frames < 0 then
						ERROR("Note has negative length", lineNum)
					end if
				else
					length32 = floor(frames / 8)				-- frames per 1/32 note
					frames = length32 * currentNote[i][2]
					if currentCutoff[i] < 0 then
						cutoffFrames = -currentCutoff[i]
						if cutoffFrames > frames then
							cutoffFrames = frames
						end if
					else				
						cutoffFrames = floor((frames * (8 - currentCutoff[i])) / 8)
					end if
					frames -= cutoffFrames --floor((frames * currentCutoff[i]) / 8)
				end if

				if frames < 1.0 * scaling then
					WARNING("Note is too short, will not be heard", lineNum)
				elsif floor(frames) > #3FFF * scaling + (scaling - 1) then
					WARNING("Note is too long, cutting at 16383 frames", lineNum)
					frames = #3FFF * scaling + (scaling - 1)
				end if

				if currentNote[i][1] = rest then
					currentNote[i][1] = CMD_REST
				else
					currentNote[i][1] = remainder(currentNote[i][1], 12)
				end if

				if pendingOctChange[i] = 1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTUP)
				elsif pendingOctChange[i] = -1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTDN)
				end if

				if useFractionalDelays then
					w1 = floor(frames) -- * 256)
					w2 = floor(cutoffFrames) -- * 256)
					if floor(frames) > 127 * scaling then
						songs[songNum][i] &= {currentNote[i][1],
								      or_bits(floor(w1 / #8000), #80),
								      and_bits(floor(w1 / #100), #7F),
								      and_bits(w1, #FF)}
					else
						songs[songNum][i] &= {currentNote[i][1],
								      floor(w1 / #100),
								      and_bits(w1, #FF)}
					end if
					if floor(cutoffFrames) > 127 *scaling then
						songs[songNum][i] &= {CMD_REST,
								      or_bits(floor(w2 / #8000), #80),
								      and_bits(floor(w2 / #100), #7F),
								      and_bits(w2, #FF)}
					elsif w2 >= 1 then
						songs[songNum][i] &= {CMD_REST, 
								      floor(w2 / #100),
								      and_bits(w2, #FF)}

					end if
				else
					w1 = floor(frames)
					w2 = floor(cutoffFrames)

					if w1 > 127 then
						songs[songNum][i] &= {currentNote[i][1],
								      or_bits(floor(w1 / 128), #80),
								      and_bits(w1, #7F)}
					elsif w1 >= 1 then
						songs[songNum][i] &= {currentNote[i][1], w1}
					end if
					if w2 > 127 then
						songs[songNum][i] &= {CMD_REST,
								      or_bits(floor(w2 / 128), #80),
								      and_bits(w2, #7F)}
					elsif w2 >= 1 then
						songs[songNum][i] &= {CMD_REST, w2}
					end if
				end if
			else
				if pendingOctChange[i] = 1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTUP)
				elsif pendingOctChange[i] = -1 then
					currentNote[i][1] = or_bits(currentNote[i][1], CMD_OCTDN)
				end if
				tuple[i] = append(tuple[i], currentNote[i])
			end if
		
			currentNote[i] = {}
			pendingOctChange[i] = 0
			hasAnyNote[i] = 1
			
		elsif forceOctChange and pendingOctChange[i] then
			songs[songNum][i] &= or_bits(CMD_OCTAVE, currentOctave[i] - minOctave)
			pendingOctChange[i] = 0
		end if
	end for
end procedure


procedure write_tuple(integer chn, sequence t, integer len)
	atom totalFrames, frames, cutoffFrames, scaling
	integer length32, w1, w2

	songLen[songNum][chn] += len
	totalFrames = (updateFreq * 60.0) / currentTempo[chn]
	if useFractionalDelays then
		length32 = floor((totalFrames / 8) * 256)
		scaling = 256
		totalFrames = length32 * len 	
		--totalFrames *= len / 8

		if currentCutoff[chn] < 0 then
			cutoffFrames = -currentCutoff[chn] * scaling
			if cutoffFrames > totalFrames then
				cutoffFrames = totalFrames
			end if
		else
			cutoffFrames = (totalFrames * (8 - currentCutoff[chn])) / 8
		end if
		totalFrames -= cutoffFrames
		if totalFrames < 0 then
			ERROR("Tuple has negative length", lineNum)
		end if
	else
	end if
	
	if totalFrames < 1.0 * scaling then
		WARNING("Note is too short, will not be heard", lineNum)
	elsif floor(totalFrames) > #3FFF * scaling + (scaling - 1) then
		WARNING("Note is too long, cutting at 16383 frames", lineNum)
		totalFrames = #3FFF * scaling + (scaling - 1)
	end if

	frames = totalFrames / length(t)
	w1 = floor(frames) -- * 256)
	for i = 1 to length(t) do
		t[i][2] = w1
	end for
	for i = 1 to floor(totalFrames) - (w1 * length(t)) do
		t[remainder(i, length(t)) + 1][2] += 1
	end for
	w2 = floor(cutoffFrames) -- * 256)
	
	for i = 1 to length(t) do
		if t[i][1] = rest then
			t[i][1] = CMD_REST
		else
			t[i][1] = remainder(t[i][1], 12)
		end if

		if useFractionalDelays then
			if floor(t[i][2] / 256) > 127 then
				songs[songNum][chn] &= {t[i][1],
						        or_bits(floor(t[i][2] / #8000), #80),
						        and_bits(floor(t[i][2] / #100), #7F),
						        and_bits(t[i][2], #FF)}
			else
				songs[songNum][chn] &= {t[i][1], floor(t[i][2] / #100), and_bits(t[i][2], #FF)}
			end if
		else
		end if
	end for
	

	if floor(cutoffFrames) > 127 * scaling then
		songs[songNum][chn] &= {CMD_REST,
				      or_bits(floor(w2 / #8000), #80),
				      and_bits(floor(w2 / #100), #7F),
				      and_bits(w2, #FF)}
	elsif w2 >= 1 then
		songs[songNum][chn] &= {CMD_REST, 
				      floor(w2 / #100),
				      and_bits(w2, #FF)}
	end if
end procedure


function get_word(integer fh)
	integer dw
	
	dw = getc(fh)
	dw += (getc(fh)*#100)
	return dw
end function


-- Get a signed word from file fh
function get_sword(integer fh)
	integer dw
	
	dw = getc(fh)
	if dw = -1 then
		return 65536
	end if
	dw += (getc(fh) * #100)
	if and_bits(dw, #8000) then
		return -(32768 - and_bits(dw, #7FFF))
	end if
	return dw
end function


-- Get a dword from file fh
function get_dword(integer fh)
	atom dw
	
	dw = getc(fh)
	dw += (getc(fh) * #100)
	dw += (getc(fh) * #10000)
	dw += (getc(fh) * #1000000)
	return dw
end function


-- Extract the sample data from a WAV file and convert it to mono with a given sample rate and volume
-- UNFINISHED
function convert_wav(sequence fname, integer sampleRate, integer volume)
	integer fn
	integer wavFormat, wavChannels, wavSamplesPerSec, wavAvgBytesPerSec,
	        wavBlockAlign, wavBitsPerSample
	sequence s, wavData
	atom dw, dataStart, dataSize
	
	fn = open(fname,"rb")
	if fn < 0 then
		ERROR("Unable to open file: " & fname, lineNum)
	end if

	s = get_bytes(fn, 4)
	if not equal(s, "RIFF") then
		ERROR("No RIFF tag found in " & fname, -1)
	end if
	
	if get_dword(fn) then end if
	
	s = get_bytes(fn,4)
	if not equal(s, "WAVE") then
		ERROR("No WAVE tag found in " & fname, -1)
	end if

	dataSize = -1
	wavData = {}
	
	-- Read the chunks
	while 1 do
		-- Get the chunk ID
		s = get_bytes(fn, 4)
		if length(s) < 4 then
			exit
		end if

		-- Get the chunk size
		dw = get_dword(fn)
	
		if equal(s, "fmt ") then
			wavFormat = get_word(fn)
			wavChannels = get_word(fn)
			wavSamplesPerSec = get_dword(fn)
			wavAvgBytesPerSec = get_dword(fn)
			wavBlockAlign = get_word(fn)
			wavBitsPerSample = get_word(fn)
		
			if wavFormat!=1 or (wavBitsPerSample!=8 and wavBitsPerSample!=16) or wavChannels > 2 then
				ERROR("Unsupported wav format in " & fname, -1)
			end if
		
		elsif equal(s, "data") then
			dataStart = where(fn)
			dataSize = dw
		end if
	end while
	
	return wavData
end function


-- Output table data in WLA-DX format 
function output_wla_table(sequence name, sequence tbl, integer canLoop, integer scaling)
	integer bytesWritten
	
	bytesWritten = 0
	
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			printf(outFile, name & "_%d:", tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						puts(outFile, ", $80")
					end if
					printf(outFile, {13, 10} & name & "_%d_loop:" & {13, 10}, tbl[1][i])
					printf(outFile, ".db $%02x, $80", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 3
				elsif j = 1 then
					printf(outFile, {13, 10} & ".db $%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				else
					printf(outFile, ", $%02x", and_bits(tbl[2][i][2][j] * scaling, #FF))
					bytesWritten += 1
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					puts(outFile, ", $80")
					bytesWritten += 1
				end if
				printf(outFile, {13, 10} & name & "_%d_loop:" & {13, 10}, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, ".db $%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					else
						printf(outFile, ", $%02x", and_bits(tbl[2][i][3][j] * scaling, #FF))
					end if
					bytesWritten += 1
				end for
				puts(outFile, ", 128")
				bytesWritten += 1
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, name & "_tbl:" & {13, 10})
		for i = 1 to length(tbl[1]) do
			printf(outFile, ".dw " & name & "_%d" & {13, 10}, tbl[1][i])
			bytesWritten += 2
		end for
		if canLoop then
			puts(outFile, name & "_loop_tbl:" & {13, 10})
			for i = 1 to length(tbl[1]) do
				printf(outFile, ".dw " & name & "_%d_loop" & {13, 10}, tbl[1][i])
				bytesWritten += 2
			end for
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, name & "_tbl:" & {13, 10})
		if canLoop then
			puts(outFile, name & "_loop_tbl:" & {13, 10})
		end if
		puts(outFile, {13, 10})
	end if
	return bytesWritten
end function


procedure output_c_table(sequence name, sequence tbl, integer canLoop)
	if length(tbl[1]) then
		for i = 1 to length(tbl[1]) do
			printf(outFile, "const unsigned char " & name & "_%d[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
			for j = 1 to length(tbl[2][i][2]) do
				if canLoop and j = length(tbl[2][i][2]) and length(tbl[2][i][3]) = 0 then
					if j > 1 then
						puts(outFile, ", 128")
					end if
					printf(outFile, "};" & {13, 10} & "const unsigned char " & name & "_%d_loop[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
					printf(outFile, "%d,128};", tbl[2][i][2][j])
				elsif j = 1 then
					printf(outFile, "%d", tbl[2][i][2][j])
				else
					printf(outFile, ",%d", tbl[2][i][2][j])
				end if
			end for
			if canLoop and length(tbl[2][i][3]) then
				if length(tbl[2][i][2]) then
					puts(outFile, ",128};")
				end if
				printf(outFile, {13, 10} & "const unsigned char " & name & "_%d_loop[] =" & {13, 10} & "{" & {13, 10}, tbl[1][i])
				for j = 1 to length(tbl[2][i][3]) do
					if j = 1 then
						printf(outFile, "%d", tbl[2][i][3][j])
					else
						printf(outFile, ",%d", tbl[2][i][3][j])
					end if
				end for
				puts(outFile, ",128};")
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, "const unsigned char *" & name & "_tbl[] =" & {13, 10} & "{" & {13, 10})
		for i = 1 to length(tbl[1]) do
			printf(outFile, name & "_%d", tbl[1][i])
			if i < length(tbl[1]) then
				puts(outFile, ",")
			end if
			puts(outFile, {13, 10})
		end for
		puts(outFile, "};" & {13, 10})
		if canLoop then
			puts(outFile, "const unsigned char *" & name & "_loop_tbl[] = " & {13, 10} & "{" & {13, 10})
			for i = 1 to length(tbl[1]) do
				printf(outFile, name & "_%d_loop", tbl[1][i])
				if i < length(tbl[1]) then
					puts(outFile, ",")
				end if
				puts(outFile, {13, 10})
			end for
			puts(outFile, "};")
		end if
		puts(outFile, {13, 10})
	else
		puts(outFile, "const unsigned char **" & name & "_tbl;" & {13, 10})
		if canLoop then
			puts(outFile, "const unsigned char **" & name & "_loop_tbl;" & {13, 10})
		end if
		puts(outFile, {13, 10})
	end if
end procedure


function ascii_to_wide(sequence s)
	sequence t
	
	t = ""
	for i = 1 to length(s) do
		t &= s[i] & 0
	end for
	
	return t
end function


function clear_bit(integer i, integer b)
	return and_bits(i, not_bits(power(2, b)))
end function


function set_bit(integer i, integer b)
	return or_bits(i, power(2, b))
end function


-- Output data suitable for the SEGA Master System / Game Gear playback library (WLA-DX)
procedure output_sms(integer writeVGM)
	atom factor, totalWaits, loopWaits
	integer f, machineSpeed, tableSize, cbSize, songSize,
	        cmd, volChange, freqChange, vol, delay
	sequence freqTbl, oct1, fileEnding, s, vgmData, channel,
	         channelDone, channelLooped, iterations, gd3,
	         cmdPos, loopPos, lastChannelSetting
	
	if writeVGM then
		fileEnding = ".vgm"
	else
		fileEnding = ".asm"
	end if

	if writeVGM then
		outFile = open(shortFilename & fileEnding, "wb")
		if outFile = -1 then
			ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
		end if

		if verbose then
			puts(1,"Generating VGM data\n")
		end if
		if updateFreq = 50 then
			machineSpeed = 3546893
		else
			machineSpeed = 3579545
		end if
		if target = TARGET_SMS and tune then
			oct1 = OCTAVE1_ALT_SMS
		else
			oct1 = OCTAVE1
		end if
		factor = 2.0
		freqTbl = repeat(0, 12 * 6)
		for i = 2 to 7 do
			for n = 1 to 12 do
				freqTbl[(i - 2) * 12 + n] = floor(machineSpeed / (oct1[n] * factor * 32))
			end for
			factor *= 2.0
		end for
		
		vgmData = {#56, #67, #6D, #20,	-- "Vgm "
		           0, 0, 0, 0,		-- EOF offset, filled in later
		           #50, #01, 0, 0}	-- VGM version (1.50)
		if updateFreq = 50 then
			vgmData &= int_to_bytes(3546893)
		else
			vgmData &= int_to_bytes(3579545)
		end if
		vgmData &= int_to_bytes(0)	-- YM2413 clock (none)
		vgmData &= int_to_bytes(0)	-- GD3 tag offset
		vgmData &= int_to_bytes(0)	-- Total samples
		vgmData &= int_to_bytes(0)	-- Loop offset
		vgmData &= int_to_bytes(0)	-- Loop samples
		vgmData &= int_to_bytes(floor(updateFreq))
		vgmData &= {9, 0, 8, 0}		-- Noise feedback pattern / shift width
		vgmData &= int_to_bytes(0)	-- YM2612 clock (none)
		vgmData &= int_to_bytes(0)	-- YM2151 clock (none)
		vgmData &= int_to_bytes(#0C)	-- Data offset
		vgmData &= {0, 0, 0, 0, 0, 0, 0, 0}
		
		totalWaits = 0
		loopWaits = 0
		channelDone = {0, 0, 0, 0}
		channelLooped = {0, 0, 0, 0}
		iterations = {0, 0}

		-- Keeps track of the numbers of waits and current size of the VGM data
		-- at the point of each command for each channel. Used for calculating
		-- the offset and length of loops in the VGM.
		cmdPos = {repeat(0, length(songs[1][1])),
		          repeat(0, length(songs[1][2])),
		          repeat(0, length(songs[1][3])),
		          repeat(0, length(songs[1][4]))}
		loopPos = {}
		
		-- Used for avoiding unnecessary writes to the PSG port.
		-- Contains the last value that have been output ({frequency, volume, panning}) for
		-- each channel.
		lastChannelSetting = repeat({-1, -1, #FF}, 4)
		
		--  1 dataPtr,
		--  2 dataPos,
		--  3 delay,
		--  4 note,
		--  5 noteOffs,
		--  6 octave,
		--  7 duty,
		--  8 freq,
		--  9 freqOffs,
		-- 10 freqOffsLatch,
		-- 11 volume,
		-- 12 volOffs,
		-- 13 volOffsLatch,
		-- 14 vMac,
		-- 15 enMac,
		-- 16 en2Mac,
		-- 17 epMac,
		-- 18 mpMac,
		-- 19 loops,
		-- 20 loopIndex,
		-- 21 detune,
		-- 22 csMac
		channel = repeat({1, 1, #100, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, {0, 0, {1, 2}}, {0, 0, {1, 2}}, {0, 0, {1, 2}}, {0, 0, {1, 2}}, {0, 0, 0}, {0, 0}, 0, 0, {0, #FF, {1, 2}}}, 4)
		while sum(channelDone) != 4 do
			for i = 1 to 4 do
				if not channelDone[i] then
					freqChange = 0
					volChange = 0
					channel[i][3] -= #100
					
					-- Check if the whole part of the delay has reached 0
					if and_bits(channel[i][3], #FFFF00) = 0 then
						iterations[2] = 0
						-- Repeat until a note command has been read
						while freqChange != 2 do
							cmd = songs[channel[i][1]][i][channel[i][2]]
							cmdPos[i][channel[i][2]] = {length(vgmData), totalWaits}
							if and_bits(cmd, #F0) = CMD_NOTE or
							   and_bits(cmd, #F0) = CMD_OCTUP or
							   and_bits(cmd, #F0) = CMD_OCTDN then
							   	-- Do the octave change if specified
							   	if and_bits(cmd, #F0) = CMD_OCTUP then
							   		channel[i][6] += 12
							   	elsif and_bits(cmd, #F0) = CMD_OCTDN then
							   		channel[i][6] -= 12
							   	end if
							   	
							   	-- If the previous note was a rest we need to trigger
							   	-- a volume change since the channel is currently muted.
							   	if channel[i][4] = CMD_REST then
							   		volChange = 1
							   	end if
								
								channel[i][4] = and_bits(cmd, #0F)
								channel[i][2] += 1
								delay = songs[channel[i][1]][i][channel[i][2]]
								if and_bits(delay, #80) then
									channel[i][2] += 1
									delay = and_bits(delay, #7F) * #80 + songs[channel[i][1]][i][channel[i][2]]
								end if
								channel[i][2] += 1
								delay = delay * #100 + songs[channel[i][1]][i][channel[i][2]]
								channel[i][3] += delay
	
								if channel[i][3] > #FF then
									freqChange = 2
								end if
								
								if channel[i][4] != CMD_REST then
									-- Reset volume macro
									if channel[i][14][1] then
										channel[i][14][3] = {1, 2}
										channel[i][14][3] = step_list(volumeMacros[2][channel[i][14][1]], channel[i][14][3])
										channel[i][11] = channel[i][14][3][3]
										volChange = 1
									end if
									-- Reset arpeggio (cumulative)
									if channel[i][15][1] then
										channel[i][15][3] = {1, 2}
										channel[i][15][3] = step_list(arpeggios[2][channel[i][15][1]], channel[i][15][3])
										channel[i][5] = channel[i][15][3][3]
									end if
									-- Reset arpeggio (non-cumulative)
									if channel[i][16][1] then
										channel[i][16][3] = {1, 2}
										channel[i][16][3] = step_list(arpeggios[2][channel[i][16][1]], channel[i][16][3])
										channel[i][5] = channel[i][16][3][3]
									end if
									-- Reset pitch sweep
									if channel[i][17][1] then
										channel[i][17][3] = {1, 2}
										channel[i][17][3] = step_list(pitchMacros[2][channel[i][17][1]], channel[i][17][3])
										channel[i][9] = channel[i][17][3][3]
									end if
									-- Reset vibrato
									if channel[i][18][1] then
										channel[i][18][3] = vibratos[2][channel[i][18][1]][2][1]
										channel[i][9] = 0
										channel[i][10] = vibratos[2][channel[i][18][1]][2][3]
									end if
									-- Reset panning
									if channel[i][22][1] and not supportsPAL then
										channel[i][22][3] = {1, 2}
										channel[i][22][3] = step_list(panMacros[2][channel[i][22][1]], channel[i][22][3])
										if channel[i][22][3][3] then
											if and_bits(channel[i][22][3][3], #80) then
												channel[i][22][2] = clear_bit(channel[i][22][2], i - 1)
												channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
											else
												channel[i][22][2] = clear_bit(channel[i][22][2], i + 3)
												channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
											end if
										else
											channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
											channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
										end if
										if channel[i][22][2] != lastChannelSetting[i][3] then
											vgmData &= {#4F, channel[i][22][2]}
											lastChannelSetting[i][3] = channel[i][22][2]
										end if
									end if
								end if
							elsif and_bits(cmd, #F0) = CMD_OCTAVE then
								channel[i][6] = and_bits(cmd, #0F) * 12
								
							elsif and_bits(cmd, #F0) = CMD_DUTY then
								channel[i][7] = and_bits(xor_bits(cmd, 1), 1) * 4

							elsif and_bits(cmd, #F0) = CMD_VOL2 then
								channel[i][11] = and_bits(cmd, #0F)
								channel[i][14][1] = 0
								volChange = 1
								
							elsif and_bits(cmd, #F0) = CMD_ARPOFF then
								channel[i][15][1] = 0
								channel[i][16][1] = 0
							
							elsif cmd = CMD_CBONCE or cmd = CMD_CBEVNT then
								channel[i][2] += 1
								
							elsif cmd = CMD_VOLMAC then
								channel[i][2] += 1
								channel[i][14][1] = songs[channel[i][1]][i][channel[i][2]]
								channel[i][14][3] = step_list(volumeMacros[2][channel[i][14][1]], {1, 2})
								channel[i][11] = channel[i][14][3][3]
								volChange = 1

							elsif cmd = CMD_PANMAC then
								channel[i][2] += 1
								channel[i][22][1] = songs[channel[i][1]][i][channel[i][2]]
								channel[i][22][3] = step_list(panMacros[2][channel[i][22][1]], {1, 2})
								if channel[i][22][3][3] then
									if and_bits(channel[i][22][3][3], #80) then
										channel[i][22][2] = clear_bit(channel[i][22][2], i - 1)
										channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
									else
										channel[i][22][2] = clear_bit(channel[i][22][2], i + 3)
										channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
									end if
								else
									channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
									channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
								end if
								if channel[i][22][2] != lastChannelSetting[i][3] and not supportsPAL then
									vgmData &= {#4F, channel[i][22][2]}
									lastChannelSetting[i][3] = channel[i][22][2]
								end if

							elsif cmd = CMD_ARPMAC then
								channel[i][2] += 1
								channel[i][15][1] = songs[channel[i][1]][i][channel[i][2]]
								channel[i][15][3] = step_list(arpeggios[2][channel[i][15][1]], {1, 2})
								channel[i][5] = channel[i][15][3][3]
								channel[i][16][1] = 0
								freqChange = 1

							elsif cmd = CMD_DUTMAC then
								channel[i][2] += 1
								
							elsif cmd = CMD_DETUNE then
								channel[i][2] += 1
								channel[i][21] = songs[channel[i][1]][i][channel[i][2]]
								
							elsif cmd = CMD_APMAC2 then
								channel[i][2] += 1
								channel[i][16][1] = songs[channel[i][1]][i][channel[i][2]]
								channel[i][16][3] = step_list(arpeggios[2][channel[i][16][1]], {1, 2})
								channel[i][15][1] = 0
								freqChange = 1
							
							elsif cmd = CMD_SWPMAC then
								channel[i][2] += 1
								channel[i][17][1] = songs[channel[i][1]][i][channel[i][2]]
								if channel[i][17][1] then
									channel[i][17][3] = step_list(pitchMacros[2][channel[i][17][1]], {1, 2})
									channel[i][9] = 0
									freqChange = 1
								else
									channel[i][9] = 0
									freqChange = 1
								end if	
							
							elsif cmd = CMD_VIBMAC then
								channel[i][2] += 1
								channel[i][18][1] = songs[channel[i][1]][i][channel[i][2]]
								if channel[i][18][1] then
									channel[i][18][3] = vibratos[2][channel[i][18][1]][2][1]
									channel[i][9] = 0
									channel[i][10] = vibratos[2][channel[i][18][1]][2][3]
									freqChange = 1
								else
									freqChange = 1
									channel[i][9] = 0
								end if
								
							elsif cmd = CMD_JMP then
								channel[i][2] = songs[channel[i][1]][i][channel[i][2] + 1] +
								                songs[channel[i][1]][i][channel[i][2] + 2] * #100
								channelLooped[i] = 1
								if sum(channelLooped) = 4 then
									loopPos = cmdPos[i][channel[i][2] + 1]
									channelDone = {1, 1, 1, 1}
									freqChange = 2
								end if
							
							elsif cmd = CMD_DJNZ then
								channel[i][19][channel[i][20]] -= 1
								if channel[i][19][channel[i][20]] != 0 then
									channel[i][2] = songs[channel[i][1]][i][channel[i][2] + 1] +
									                songs[channel[i][1]][i][channel[i][2] + 2] * #100
								else
									channel[i][20] -= 1
									channel[i][2] += 2
								end if
								
							elsif cmd = CMD_LOPCNT then
								channel[i][2] += 1
								channel[i][20] += 1
								channel[i][19][channel[i][20]] = songs[channel[i][1]][i][channel[i][2]]
								
							elsif cmd = CMD_END then
								channel[i][4] = cmd
								channelDone[i] = 1
								freqChange = 2
								
							end if
							
							channel[i][2] += 1
							iterations[2] += 1
							if iterations[2] = 131072 then
								puts(1, "Internal error: Appears to be stuck in an infinite loop. Exiting\n")
								? {i, channel[1][2], channel[2][2], channel[3][2], channel[4][2]}
								exit
							end if
						end while
					else
						--channel[i][3] -= #100
						
						-- Update effects as needed
						if channel[i][14][1] then
							channel[i][14][3] = step_list(volumeMacros[2][channel[i][14][1]], channel[i][14][3])
							channel[i][11] = channel[i][14][3][3]
							volChange = 1
						end if
						if channel[i][15][1] then
							channel[i][15][3] = step_list(arpeggios[2][channel[i][15][1]], channel[i][15][3])
							channel[i][5] += channel[i][15][3][3]
							freqChange = 1
						end if
						if channel[i][16][1] then
							channel[i][16][3] = step_list(arpeggios[2][channel[i][16][1]], channel[i][16][3])
							channel[i][5] = channel[i][16][3][3]
							freqChange = 1
						end if
						if channel[i][17][1] then
							channel[i][17][3] = step_list(pitchMacros[2][channel[i][17][1]], channel[i][17][3])
							channel[i][9] += channel[i][17][3][3]
							freqChange = 1
						end if
						if channel[i][18][1] then
							if channel[i][18][3] = 0 then
								channel[i][9] = channel[i][10]
								channel[i][10] = -channel[i][10]
								channel[i][18][3] = vibratos[2][channel[i][18][1]][2][2]
								freqChange = 1
							end if
							channel[i][18][3] -= 1
						end if
						if channel[i][22][1] and not supportsPAL then
							channel[i][22][3] = step_list(panMacros[2][channel[i][22][1]], channel[i][22][3])
							if channel[i][22][3][3] then
								if and_bits(channel[i][22][3][3], #80) then
									channel[i][22][2] = clear_bit(channel[i][22][2], i - 1)
									channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
								else
									channel[i][22][2] = clear_bit(channel[i][22][2], i + 3)
									channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
								end if
							else
								channel[i][22][2] = set_bit(channel[i][22][2], i - 1)
								channel[i][22][2] = set_bit(channel[i][22][2], i + 3)
							end if
							if channel[i][22][2] != lastChannelSetting[i][3] then
								vgmData &= {#4F, channel[i][22][2]}
								lastChannelSetting[i][3] = channel[i][22][2]
							end if
						end if
					end if
					if channel[i][4] = CMD_END or channel[i][4] = CMD_REST then
						if lastChannelSetting[i][2] != 0 then
							vgmData &= {#50, or_bits(#9F, (i - 1) * #20)}
							vgmData &= {#50, or_bits(#1F, (i - 1) * #20)}
							lastChannelSetting[i] = {-1, 0, lastChannelSetting[i][3]}
						end if
						if channel[i][4] = CMD_END then
							channelDone[i] = 1
							channelLooped[i] = 1
						end if
					else	
						if freqChange then
							if i < 4 then
								if channel[i][6] + channel[i][4] + channel[i][5] < 0 then
									channel[i][8] = freqTbl[1]
								elsif channel[i][6] + channel[i][4] + channel[i][5] > 59 then
									channel[i][8] = freqTbl[60]
								else
									channel[i][8] = freqTbl[channel[i][6] + channel[i][4] + channel[i][5] + 1]
								end if
								channel[i][8] -= channel[i][9] + channel[i][21]
				
								if channel[i][8] >= #03EF then
									channel[i][8] = #03EF
								elsif channel[i][8] < #001C then
									channel[i][8] = #001C
								end if
								if channel[i][8] != lastChannelSetting[i][1] then
									vgmData &= {#50, or_bits(#80 + (i - 1) * #20, and_bits(channel[i][8], #0F))}
									vgmData &= {#50, floor(and_bits(channel[i][8], #3F0) / #10)}
									lastChannelSetting[i][1] = channel[i][8]
								end if
							else
								vgmData &= {#50, #E0 + channel[i][7] + and_bits(channel[i][4], 3)}
								vgmData &= {#50, #60 + channel[i][7] + and_bits(channel[i][4], 3)}
							end if
						end if
						if volChange then
							vol = channel[i][11] + channel[i][12]
							if vol > 15 then
								vol = 15
							elsif vol < 0 then
								vol = 0
							end if
							if vol != lastChannelSetting[i][2] then
								vgmData &= {#50, or_bits(#90 + xor_bits(vol, 15), (i - 1) * #20)}
								vgmData &= {#50, or_bits(#10 + xor_bits(vol, 15), (i - 1) * #20)}
								lastChannelSetting[i][2] = vol
							end if
						end if
					end if
				end if
			end for
			-- Wait one frame
			if updateFreq = 50 then
				vgmData &= #63
				totalWaits += 882
			else
				vgmData &= #62
				totalWaits += 735
			end if
			iterations[1] += 1
			if iterations[1] = 54000 then
				puts(1, "Internal error: Appears to be stuck in an infinite loop. Exiting\n")
				? {channel[1][2], channel[2][2], channel[3][2], channel[4][2]}
				exit
			end if
		end while
		if length(loopPos) then
			vgmData = vgmData[1..#18] &
			          int_to_bytes(totalWaits) &
			          int_to_bytes(loopPos[1] - #1C) &
			          int_to_bytes(totalWaits - loopPos[2]) &
			          vgmData[#25..length(vgmData)]
			loopPos[2] = totalWaits - loopPos[2]
		else
			vgmData = vgmData[1..#18] & int_to_bytes(totalWaits) & vgmData[#1D..length(vgmData)]
			loopPos = {0, 0}
		end if	

		vgmData &= #66
		
		vgmData = vgmData[1..4] & int_to_bytes(length(vgmData) - 4) & vgmData[9..length(vgmData)]
		vgmData = vgmData[1..#14] & int_to_bytes(length(vgmData) - #14) & vgmData[#19..length(vgmData)]
		puts(outFile, vgmData)
		
		s = date()
		gd3 = ascii_to_wide(songTitle & 0 & 0)
		gd3 &= ascii_to_wide("Unknown" & 0 & 0)
		if supportsPAL then
			gd3 &= ascii_to_wide("Sega Master System" & 0 & 0)
		else
			gd3 &= ascii_to_wide("Sega Game Gear" & 0 & 0)
		end if
		gd3 &= ascii_to_wide(songProgrammer & 0 & 0)
		gd3 &= ascii_to_wide(sprintf("%04d/%02d/%02d", {s[1] + 1900, s[2], s[3]}) & 0)
		gd3 &= ascii_to_wide("XPMC" & 0)
		gd3 &= ascii_to_wide("Composer: " & songComposer & 0)
		gd3 = "Gd3 " &
		      {0, 1, 0, 0} &
		      int_to_bytes(length(gd3)) &
		      gd3
		puts(outFile, gd3)
		close(outFile)

		if verbose then
			for i = 1 to 4 do
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d / %d ticks\n", {1, songLen[1][i], songLoopLen[1][i]})
			end for
			printf(1,"VGM size: %d bytes + %d bytes GD3\nVGM length: %d / %d seconds\n", {length(vgmData), length(gd3), floor(totalWaits / 44100), floor(loopPos[2] / 44100)})
		end if
     	
	else
		outFile = open(shortFilename & fileEnding, "wb")
		if outFile = -1 then
			ERROR("Unable to open file: " & shortFilename & fileEnding, -1)
		end if

		s = date()
		printf(outFile, "; Written by XPMC at %02d:%02d:%02d on " & WEEKDAYS[s[7]] & " " & MONTHS[s[2]] & " %d, %d." & {13, 10, 13, 10},
		       s[4..6] & {s[3], s[1] + 1900})
		       
		if not supportsPAL then
			puts(outFile, ".DEFINE XPMP_GAME_GEAR" & {13, 10})
		end if

		if updateFreq = 50 then
			puts(outFile, ".DEFINE XPMP_50_HZ" & {13, 10})
			machineSpeed = 3546893
		else
			machineSpeed = 3579545
		end if

		if target = TARGET_SMS and tune then
			puts(outFile, ".DEFINE XPMP_TUNE_SMS" & {13, 10})
		end if

		if not length(volumeMacros[1]) then
			puts(outFile, ".DEFINE XPMP_VMAC_NOT_USED" & {13, 10})
		end if
		if not length(pitchMacros[1]) then
			puts(outFile, ".DEFINE XPMP_EPMAC_NOT_USED" & {13, 10})
		end if
		if not length(vibratos[1]) then
			puts(outFile, ".DEFINE XPMP_MPMAC_NOT_USED" & {13, 10})
		end if
		if not usesEN[1] then
			puts(outFile, ".DEFINE XPMP_ENMAC_NOT_USED" & {13, 10})
		end if
		if not usesEN[2] then
			puts(outFile, ".DEFINE XPMP_EN2MAC_NOT_USED" & {13, 10})
		end if
		
		-- Write frequency table
		--factor = 2.0
		--for i = 2 to 7 do
		--	puts(outFile, ".dw ")
		--	for n = 1 to 12 do
		--		f = floor(machineSpeed / (OCTAVE1_ALT_SMS[n] * factor * 32))
		--		printf(outFile, "$%04x", and_bits(f, #3F0) * 16 + and_bits(f, #0F) + #80)
		--		if n < 12 then
		--			puts(outFile, ",")
		--		end if
		--	end for
		--	puts(outFile, {13,10})
		--	factor *= 2.0
		--end for
		--puts(outFile, {13, 10})

		tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1)
		tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1)
		tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1)
		tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1)
		tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1)
		tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1)
		tableSize += output_wla_table("xpmp_CS_mac", panMacros,    1, 1)

		if verbose then
			printf(1, "Size of effect tables: %d bytes\n", tableSize)
		end if

		cbSize = 0
		puts(outFile, "xpmp_callback_tbl:" & {13, 10})
		for i = 1 to length(callbacks) do
			puts(outFile, ".dw " & callbacks[i] & {13, 10})
			cbSize += 2
		end for
		puts(outFile, {13, 10})

		if verbose then
			printf(1, "Size of callback table: %d bytes\n", cbSize)
		end if

		songSize = 0
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels) do
					printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
					for j = 1 to length(songs[n][i]) do
						if remainder(j, 16) = 1 then
							puts(outFile, {13, 10} & ".db ")
						end if				
						printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
						if j < length(songs[n][i]) and remainder(j, 16) != 0 then
							puts(outFile, ",")
						end if

					end for
					puts(outFile, {13, 10})
					printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), songLen[n][i], songLoopLen[n][i]})
					songSize += length(songs[n][i])
				end for
			end if
		end for

		puts(outFile, {13, 10} & "xpmp_song_tbl:" & {13, 10})
		for n = 1 to length(songs) do
			if sequence(songs[n]) then
				for i = 1 to length(supportedChannels) do
					printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & {13, 10}, n)
					songSize += 2
				end for
			end if
		end for

		if verbose then
			printf(1, "Total size of song(s): %d bytes\n", songSize + tableSize + cbSize)
		end if

		close(outFile)
	end if
end procedure


-- Output data suitable for the Gameboy / Gameboy Color
procedure output_gbc()
	atom factor,f2,r2,smallestDiff
	integer f, tableSize
	sequence closest

	outFile = open(shortFilename & ".asm", "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & ".asm", -1)
	end if

	tableSize  = output_wla_table("xpmp_dt_mac", dutyMacros,   1, 1)
	tableSize += output_wla_table("xpmp_v_mac",  volumeMacros, 1, 1)
	tableSize += output_wla_table("xpmp_VS_mac", volumeSlides, 1, 1)
	tableSize += output_wla_table("xpmp_EP_mac", pitchMacros,  1, 1)
	tableSize += output_wla_table("xpmp_EN_mac", arpeggios,    1, 1)
	tableSize += output_wla_table("xpmp_MP_mac", vibratos,     0, 1)
	
	puts(outFile, "xpmp_waveform_data:")
	for i = 1 to length(waveforms[2]) do
		for j = 1 to length(waveforms[2][i]) by 2 do
			if j = 1 then
				puts(outFile, {13, 10} & ".db ")
			end if				
			printf(outFile, "$%02x", waveforms[2][i][j]*#10 + waveforms[2][i][j+1])
			if j < length(waveforms[2][i]) - 1 then
				puts(outFile, ",")
			end if

		end for
	end for
	puts(outFile, {13, 10, 13, 10})
				
	puts(outFile, "xpmp_callback_tbl:" & {13, 10})
	for i = 1 to length(callbacks) do
		puts(outFile, ".dw " & callbacks[i] & {13, 10})
	end for
	puts(outFile, {13, 10})
	
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i] & ":", n)
				for j = 1 to length(songs[n][i]) do
					if remainder(j, 16) = 1 then
						puts(outFile, {13, 10} & ".db ")
					end if				
					printf(outFile, "$%02x", and_bits(songs[n][i][j], #FF))
					if j < length(songs[n][i]) and remainder(j, 16) != 0 then
						puts(outFile, ",")
					end if

				end for
				puts(outFile, {13, 10})
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), songLen[n][i], songLoopLen[n][i]})
			end for
		end if
	end for
	
	puts(outFile, {13, 10} & "xpmp_song_tbl:" & {13, 10})
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				printf(outFile, ".dw xpmp_s%d_channel_" & supportedChannels[i] & {13, 10}, n)
			end for
		end if
	end for
	
	close(outFile)
end procedure


procedure output_nds()
	integer first
	
	outFile = open(shortFilename & ".h", "wb")
	if outFile = -1 then
		ERROR("Unable to open file: " & shortFilename & ".h", -1)
	end if

	output_c_table("xpmp_dt_mac", dutyMacros,   1)
	output_c_table("xpmp_v_mac",  volumeMacros, 1)
	output_c_table("xpmp_VS_mac", volumeSlides, 1)
	output_c_table("xpmp_EP_mac", pitchMacros,  1)
	output_c_table("xpmp_EN_mac", arpeggios,    1)
	output_c_table("xpmp_MP_mac", vibratos,     0)
	
	puts(outFile, "const xpmp_callback xpmp_callback_tbl[] =" & {13, 10} & "{" & {13, 10})
	for i = 1 to length(callbacks) do
		puts(outFile, callbacks[i])
		puts(outFile, ",")
		puts(outFile, {13, 10})
	end for
	puts(outFile, "NULL" & {13, 10} & "};" & {13, 10})
	
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				printf(outFile, "const unsigned char xpmp_s%d_channel_" & supportedChannels[i] & " = " & {13,10} & "{" & {13, 10}, n)
				for j = 1 to length(songs[n][i]) do
					printf(outFile, "0x%02x", songs[n][i][j])
					if j < length(songs[n][i]) then
						puts(outFile, ",")
					end if
					if remainder(j, 16) = 0 then
						puts(outFile, {13, 10})
					end if
				end for
				puts(outFile, "};" & {13, 10})
				printf(1, "Song %d, Channel " & supportedChannels[i] & ": %d bytes, %d / %d ticks\n", {n, length(songs[n][i]), songLen[n][i], songLoopLen[n][i]})
			end for
		end if
	end for
	
	puts(outFile, {13, 10} & "unsigned char *xpmp_song_tbl[] = " & {13, 10} & "{" & {13, 10})
	first = 1
	for n = 1 to length(songs) do
		if sequence(songs[n]) then
			for i = 1 to length(supportedChannels) do
				if first then
					first = 0
				else
					puts(outFile, "," & {13,10})
				end if
				printf(outFile, "xpmp_s%d_channel_" & supportedChannels[i], n)
			end for
		end if
	end for
	puts(outFile, {13, 10} & "};" & {13, 10})
	
	close(outFile)
end procedure


procedure compile(sequence fileNames)
	integer idx, l, m, n, n2, numChannels,
	        lastWasChannelSelect, prevLine, bypass,
	        note, firstNote, flatSharp, len, dot, tie, slur,
	        hasTie, hasSlur, writeVGM
	sequence s, t, u, macro, dontCompile
	object o
	
	songTitle 	= "Unknown"
	songComposer 	= "Unknown"
	songProgrammer 	= "Unknown"
	
	minNote 	= 1
	supportsPAL 	= 0
	useFractionalDelays = 1
	writeVGM = 0
	
	defines = {{}, {}}
	
	tune = 0
	
	if target = TARGET_SMS then
		define("SMS", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D'}
		supportsDutyChange 	= {-1, -1, -1, 1}
		supportsVolumeChange 	= {1, 1, 1, 1}
		supportsWave 		= {0, 0, 0, 0}
		supportsPCM 		= {0, 0, 0, 0}
		activeChannels 		= {0, 0, 0, 0}
		supportsHwToneEnv	= {0, 0, 0, 0}
		supportsHwVolEnv	= {0, 0, 0, 0}
		supportsDetune		= {1, 1, 1, 0}
		maxTempo 		= 250
		minOctave 		= 2
		maxOctave 		= 7
		minVolume 		= 0
		maxVolume 		= {15, 15, 15, 15}
		supportsPan 		= 0
		minNote 		= 10		-- Lowest possible note is A in octave 2
		maxLoopDepth 		= 2
		supportsPAL 		= 1
	
	elsif target = TARGET_SGG then
		define("SGG", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D'}
		supportsDutyChange 	= {-1, -1, -1, 1}
		supportsVolumeChange 	= {1, 1, 1, 1}
		supportsWave 		= {0, 0, 0, 0}
		supportsPCM 		= {0, 0, 0, 0}
		activeChannels 		= {0, 0, 0, 0}
		supportsHwToneEnv	= {0, 0, 0, 0}
		supportsHwVolEnv	= {0, 0, 0, 0}
		supportsDetune		= {1, 1, 1, 0}
		maxTempo 		= 250
		minOctave 		= 2
		maxOctave 		= 7
		minVolume 		= 0
		maxVolume 		= {15, 15, 15, 15}
		supportsPan 		= 1
		minNote 		= 10		-- Lowest possible note is A in octave 2
		maxLoopDepth 		= 2

	elsif target = TARGET_GBC then
		define("GBC", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D'}
		supportsDutyChange 	= {3, 3, -1, 1}
		supportsVolumeChange 	= {1, 1, 1, 1}
		supportsWave 		= {0, 0, 1, 0}
		supportsPCM 		= {0, 0, 0, 0}
		activeChannels 		= {0, 0, 0, 0}
		supportsHwToneEnv	= {7, 0, 0, 0}
		supportsHwVolEnv	= {7, 7, 0, 7}
		supportsDetune		= {1, 1, 0, 0}
		maxTempo 		= 250
		minOctave 		= 2
		maxOctave 		= 7
		minVolume 		= 0
		maxVolume 		= {15, 15, 3, 15}
		supportsPan 		= 0
		maxLoopDepth 		= 2
		maxWavLength		= 32
		minWavSample 		= 0
		maxWavSample		= 15
		
	elsif target = TARGET_NDS then
		define("NDS", 1)
		supportedChannels 	= {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J'}
		supportsDutyChange 	= {7, 7, 7, 7, -1, -1, -1, -1, -1, -1}
		supportsVolumeChange 	= {1, 1, 1, 1, 1, 1, 1, 1, 1, 1}
		supportsWave 		= {0, 0, 0, 0, 0, 0, 1, 1, 1, 1}
		supportsPCM 		= {0, 0, 0, 0, 0, 0, 1, 1, 1, 1}
		activeChannels 		= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
		supportsHwToneEnv	= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
		supportsHwVolEnv	= {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
		supportsDetune 		= {1, 1, 1, 1, 1, 1, 1, 1, 1, 1}
		maxTempo 		= 250
		minOctave 		= 2
		maxOctave 		= 7
		minVolume 		= 0
		maxVolume 		= {127, 127, 127, 127, 127, 127, 127, 127, 127, 127}
		supportsPan 		= 1
		maxLoopDepth 		= 4
		maxWavLength		= 262144
		minWavSample 		= -128
		maxWavSample		= 127
		
	elsif target = TARGET_XGS then
		define("XGS", 1)
		supportedChannels 	= {'A', 'B', 'C'}
		supportsDutyChange 	= {-1, -1, -1}
		supportsVolumeChange 	= {1, 1, 1}
		supportsWave 		= {0, 0, 0}
		supportsPCM 		= {0, 0, 0}
		activeChannels 		= {0, 0, 0}
		supportsHwToneEnv	= {0, 0, 0}
		supportsHwVolEnv	= {0, 0, 0}
		supportsDetune		= {0, 0, 0}
		maxTempo 		= 250
		minOctave 		= 2
		maxOctave 		= 7
		minVolume 		= 0
		maxVolume 		= {15, 15, 15}
		supportsPan 		= 1
		maxLoopDepth 		= 2
		useFractionalDelays	= 0
		
	else
		ERROR("No target specified", -1)
	end if

	machineVolLimit = maxVolume
	rest = (maxOctave + 1 - minOctave) * 12
	
	n = length(fileNames[1])
	while n > 0 and fileNames[1][n] != '.' do
		n -= 1
	end while
	if n = 0 then
		shortFilename = fileNames[1]
		fileNames[1] &= ".mml"
	else
		shortFilename = fileNames[1][1..n - 1]
	end if
	if length(fileNames) > 1 then
		shortFilename = fileNames[2]
		n = length(fileNames[2])
		while n > 0 and fileNames[2][n] != '.' do
			n -= 1
		end while
		if n != 0 then
			shortFilename = fileNames[2][1..n - 1]
			writeVGM = equal(lower(fileNames[2][n..length(fileNames[2])]), ".vgm")
		end if
	end if
	inFile = open(fileNames[1], "rb")
	if inFile = -1 then
		ERROR("Unable to open file: " & fileNames[1], -1)
	end if
	
	n = 0
	fileData = ""
	while n != -1 do
		n = getc(inFile)
		if n != -1 then
			fileData &= n
		end if
	end while
	close(inFile)
	fileDataPos = 1
	fileData &= -1
	
	songs 		= repeat(0, 100)
	songLen 	= repeat(0, 100)
	songs[1] 	= repeat({}, length(supportedChannels))
	songLen[1] 	= repeat(0, length(supportedChannels))

	songLoopLen 	= songLen
	songNum 	= 1
	lineNum 	= 1
	prevLine 	= 1
	updateFreq 	= 60.0
	tie 		= 0
	slur 		= 0
	octaveRev 	= 1
	enRev 		= 0
	lastWasChannelSelect = 0
	
	dontCompile	= {0}
	macros 		= {{}, {}}
	volumeMacros 	= {{}, {}}
	panMacros 	= {{}, {}}
	volumeSlides 	= {{}, {}}
	dutyMacros 	= {{}, {}}
	pitchMacros 	= {{}, {}}
	arpeggios 	= {{}, {}}
	vibratos 	= {{}, {}}
	portamentos 	= {{}, {}}
	waveforms	= {{}, {}}
	pcms		= {{}, {}}
	callbacks 	= {}
	loopStack 	= repeat({}, length(supportedChannels))
	currentLength 	= repeat(8, length(supportedChannels))
	currentOctave 	= repeat(4, length(supportedChannels))
	currentTempo 	= repeat(125, length(supportedChannels))
	currentCutoff 	= repeat(8, length(supportedChannels))
	currentVolume 	= floor(maxVolume / 2)
	loopPoint 	= repeat(-1, length(supportedChannels))
	currentNote 	= repeat({}, length(supportedChannels))
	lastSetLength 	= repeat(0, length(supportedChannels))
	pendingOctChange = repeat(0, length(supportedChannels))
	hasAnyNote    	= repeat(0, length(supportedChannels))
	tuple		= repeat(0, length(supportedChannels))
	usesEN		= {0, 0}
	
	while 1 do
		bypass = 0
		
		n = getch()
		if n = -1 then
			exit
		end if
		
		n2 = n
		
		if n = 10 then
			lineNum += 1
		end if
		
		if lineNum > prevLine then
			activeChannels = repeat(0, length(supportedChannels))
			prevLine = lineNum
		end if
		
		if dontCompile[1] then
			if n = '#' then
				s = get_string()
				if equal(s, "IFDEF") then
					s = get_string_until({13, 10})
					dontCompile = or_bits((not is_defined(s)), dontCompile[1]) & dontCompile

				elsif equal(s, "IFNDEF") then
					s = get_string_until({13, 10})
					dontCompile = or_bits(is_defined(s), dontCompile[1]) & dontCompile

				elsif equal(s, "ELSIFDEF") then
					if length(dontCompile) > 1 then
						s = get_string_until({13, 10})
						dontCompile[1] = or_bits((not is_defined(s)), dontCompile[2])
					else
						ERROR("ELSIFDEF with no matching IFDEF", lineNum)
					end if
					
				elsif equal(s, "ELSE") then
					if length(dontCompile) > 1 then
						dontCompile[1] = or_bits((not dontCompile[1]), dontCompile[2])
					else
						ERROR("ELSE with no matching IFDEF", lineNum)
					end if

				elsif equal(s, "ENDIF") then
					if length(dontCompile) > 1 then
						dontCompile = dontCompile[2..length(dontCompile)]
					else
						ERROR("ENDIF with no matching IFDEF", lineNum)
					end if
				end if
			end if
		else
		
			if n = '#' then
				write_note(1)
				s = get_string()
				if equal(s, "TITLE") then
					songTitle = get_string_until({13, 10})

				elsif equal(s, "TUNE") then
					tune = 1
					
				elsif equal(s, "COMPOSER") then
					songComposer = get_string_until({13, 10})

				elsif equal(s, "PROGRAMER") or equal(s, "PROGRAMMER") then
					songProgrammer = get_string_until({13, 10})

				elsif equal(s, "INCLUDE") then
					skip_whitespace()
					if getch() = '\"' then
						s = get_string_until({'\"'})
						if getch() = '\"' then
							inFile = open(s, "rb")
							if inFile = -1 then
								ERROR("Unable to open file: " & s, lineNum)
							end if
							m = 0
							s = ""
							while m != -1 do
								m = getc(inFile)
								if m != -1 then
									s &= m
								end if
							end while
							close(inFile)
							if fileDataPos <= length(fileData) then
								fileData = fileData[1..fileDataPos - 1] & s & fileData[fileDataPos..length(fileData)]
							else
								fileData &= s
							end if
						else
							ERROR("Malformed #INCLUDE, missing ending \"", lineNum)
						end if
					else
						ERROR("Malformed #INCLUDE, missing starting \"", lineNum)
					end if
					
				elsif equal(s, "PAL") then
					if supportsPAL then
						updateFreq = 50.0
					end if

				elsif equal(s, "NTSC") then
					updateFreq = 60.0

				elsif equal(s, "IFDEF") then
					s = get_string_until({13, 10})
					dontCompile = or_bits((not is_defined(s)), dontCompile[1]) & dontCompile

				elsif equal(s, "IFNDEF") then
					s = get_string_until({13, 10})
					dontCompile = or_bits(is_defined(s), dontCompile[1]) & dontCompile

				elsif equal(s, "ELSIFDEF") then
					if length(dontCompile) > 1 then
						s = get_string_until({13, 10})
						dontCompile[1] = or_bits((not is_defined(s)), dontCompile[2])
					else
						ERROR("ELSIFDEN with no matching IFDEF", lineNum)
					end if
					
				elsif equal(s, "ELSE") then
					if length(dontCompile) > 1 then
						dontCompile[1] = or_bits((not dontCompile[1]), dontCompile[2])
					else
						ERROR("ELSE with no matching IFDEF", lineNum)
					end if

				elsif equal(s, "ENDIF") then
					if length(dontCompile) > 1 then
						dontCompile = dontCompile[2..length(dontCompile)]
					else
						ERROR("ENDIF with no matching IFDEF", lineNum)
					end if

				elsif equal(s, "SONG") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if integer(o[2]) and o[2] > 1 and o[2] < 100 then
							if integer(songs[o[2]]) then
								m = -1
								l = 0
								for i = 1 to length(supportedChannels) do
									if songLen[songNum][i] then
										if m = -1 then
											m = songLen[songNum][i]
										elsif songLen[songNum][i] != m then
											WARNING("Mismatch in length between channels", -1)
											exit
										end if
									end if
									songLoopLen[songNum][i] = songLen[songNum][i] - songLoopLen[songNum][i]
									if loopPoint[i] = -1 then
										songs[songNum][i] &= CMD_END
									else
										if not hasAnyNote[i] then
											songs[songNum][i] &= CMD_END
										else
											songs[songNum][i] &= {CMD_JMP, and_bits(loopPoint[i], #FF), floor(loopPoint[i] / #100)}
										end if
									end if
									loopPoint[i] = -1
								end for
								songNum = o[2]
								songs[songNum] = repeat({}, length(supportedChannels))
								songLen[songNum] = repeat(0, length(supportedChannels))
								hasAnyNote = repeat(0, length(supportedChannels))
								songLoopLen[songNum] = songLen[songNum]
								for i = 1 to length(loopStack) do
									if length(loopStack[i]) then
										ERROR("Open [ on channel " & supportedChannels[i], lineNum)
									end if
								end for
							else
								ERROR("Song " & s & " already defined", lineNum)
							end if
						else
							ERROR("Bad song number: " & s, lineNum)
						end if
					else
						ERROR("Bad song number: " & s, lineNum)
					end if

				elsif equal(s, "UNIFORM-VOLUME") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if o[2] > 1 then
							maxVolume = repeat(o[2], length(supportedChannels))
							if o[2] > 255 then
								WARNING("Very large max volume specified: " & s, lineNum)
							end if
						else
							ERROR("Volume must be >= 1: " & s, lineNum)
						end if
					else
						ERROR("Expected a positive number: " & s, lineNum)
					end if

				elsif equal(s, "EN-REV") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if o[2] = 1 then
							enRev = 1
						elsif o[2] = 0 then
						else
							WARNING("Expected 0 or 1, defaulting to 0: " & s, lineNum)
						end if
					else
						ERROR("Expected 0 or 1: " & s, lineNum)
					end if

				elsif equal(s, "OCTAVE-REV") then
					s = get_string()
					o = value(s)
					if o[1] = GET_SUCCESS then
						if o[2] = 1 then
							octaveRev = -1
						elsif o[2] = 0 then
						else
							WARNING("Expected 0 or 1, defaulting to 0: " & s, lineNum)
						end if
					else
						ERROR("Expected 0 or 1: " & s, lineNum)
					end if

				elsif equal(s, "AUTO-BANKSWITCH") then
					WARNING("Unsupported command: AUTO-BANKSWITCH", lineNum)
					s = get_string() 

				else
					ERROR("Unknown command: " & s, lineNum)
				end if


			elsif n = '@' then
				write_note(1)
				m = getch()
				if m = '@' then
					s = "@" & get_numeric_string()
				elsif is_numeric(m) then
					ungetch()
					s = get_numeric_string()
				else
					l = getch()
					if is_alpha(l) then
						s = {m, l} & get_numeric_string()
					else
						ungetch()
						s = {m} & get_numeric_string()
					end if
				end if

				if length(s) then
					if s[1] = '@' then
						o = value(substr(s, 2, length(s)))
						if o[1] = GET_SUCCESS then
							--? o[2]
							idx = find(o[2], dutyMacros[1])
							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
									if supportsDutyChange[i] then
										songs[songNum][i] &= {CMD_DUTMAC, idx}
									else
										WARNING("Unsupported command for channel " &
											supportedChannels[i] & ": @@", lineNum)
									end if
								end if
							end for
							if numChannels = 0 then
								WARNING("Use of @@ with no channels active", lineNum)
							end if
						else
							ERROR("Expected a number: " & s, lineNum)
						end if
					else
						o = value(s)
						-- Duty cycle macro definition
						if o[1] = GET_SUCCESS then
							idx = find(o[2], dutyMacros[1])
							if sum(activeChannels) = 0 then
								if idx <= 0 then
									dutyMacros[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) or length(t[3]) then
												dutyMacros[2] = append(dutyMacros[2], t)
											else
												ERROR("Empty sequence", lineNum)
											end if
										else
											ERROR("Syntax error", lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								numChannels = 0
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										numChannels += 1
										if o[2] >= 0 and o[2] <= supportsDutyChange[i] then
											songs[songNum][i] &= or_bits(CMD_DUTY, o[2])
										else
											WARNING("Unsupported command for channel "& 
												supportedChannels[i] & ": @", lineNum)
										end if
									end if
								end for
								if numChannels = 0 then
									WARNING("Use of @ with no channels active", lineNum)
								end if
							end if

						-- Pan macro definition
						elsif equal(substr(s, 1, 2), "CS") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], panMacros[1])
								if idx <= 0 then
									panMacros[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) or length(t[3]) then
												if in_range(t[2] & t[3], -63, 63) then
													panMacros[2] = append(panMacros[2], t)
												else
													ERROR("Value of out range (allowed: -63-63): " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Empty sequence", lineNum)
											end if
										else
											ERROR("Bad CS: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if			
							
						-- Arpeggio definition
						elsif equal(substr(s, 1, 2), "EN") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], arpeggios[1])
								if idx <= 0 then
									arpeggios[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) or length(t[3]) then
												if in_range(t[2] & t[3], -63, 63) then
													arpeggios[2] = append(arpeggios[2], t)
												else
													ERROR("Value of out range (allowed: -63-63): " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Empty sequence", lineNum)
											end if
										else
											ERROR("Bad EN: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if						

						-- Vibrato definition
						elsif equal(substr(s, 1, 2), "MP") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], vibratos[1])
								if idx <= 0 then
									vibratos[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) = 3 and length(t[3]) = 0 then
												if in_range(t[2], {0, 1, 0}, {127, 127, 63}) then
													vibratos[2] = append(vibratos[2], t)
												else
													ERROR("Value of out range: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad MP: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad MP: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if						

						-- Pitch macro definition
						elsif equal(substr(s, 1, 2), "EP") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], pitchMacros[1])
								if idx <= 0 then
									pitchMacros[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											pitchMacros[2] = append(pitchMacros[2], t)
										else
											ERROR("Bad EP: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		

						-- Portamento definition
						elsif equal(substr(s, 1, 2), "PT") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], portamentos[1])
								if idx <= 0 then
									portamentos[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) = 2 and length(t[3]) = 0 then
												if in_range(t[2], {0, 1}, {127, 127}) then
													portamentos[2] = append(portamentos[2], t)
												else
													ERROR("Value out of range: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad PT: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad PT: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		

						-- Waveform definition
						elsif equal(substr(s, 1, 2), "WT") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], waveforms[1])
								if idx <= 0 then
									waveforms[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if sum(supportsWave) != 0 then
											if t[1] = GET_SUCCESS then
												if length(t[3]) = 0 then
													if in_range(t[2], minWavSample, maxWavSample) then
														if length(t[2]) < maxWavLength then
															WARNING("Padding waveform with zeroes", lineNum)
															t[2] &= repeat(0, maxWavLength - length(t[2]))
														elsif length(t[2]) > maxWavLength then
															WARNING("Truncating waveform", lineNum)
															t[2] = t[2][1..maxWavLength]
														end if
														waveforms[2] = append(waveforms[2], t[2])
													else
														ERROR("Waveform data out of range: " & sprint_list(t), lineNum)
													end if
												else
													ERROR("Loops not supported in waveform: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad waveform: " & t, lineNum)
											end if
										else
											WARNING("Unsupported command for this target: @WT", lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		
							
						-- Volume slide definition
						elsif equal(substr(s, 1, 2), "VS") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], volumeSlides[1])
								if idx <= 0 then
									volumeSlides[1] &= o[2]
									t = get_string()
									if equal(t, "=") then
										t = get_list()
										if t[1] = GET_SUCCESS then
											if length(t[2]) = 3 and length(t[3]) = 0 then
												if in_range(t[2], {0, 1, -63}, {127, 127, 63}) then									
													volumeSlides[2] = append(volumeSlides[2], t)
												else
													ERROR("Bad VS: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad VS: " & sprint_list(t), lineNum)
											end if
										else
											ERROR("Bad VS: " & t, lineNum)
										end if
									else
										ERROR("Expected \'=\'", lineNum)
									end if
								else
									ERROR("Redefinition of @" & s, lineNum)
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		

						-- Volume macro definition
						elsif equal(substr(s, 1, 1), "v") then
							o = value(substr(s, 2, length(s)))
							if o[1] = GET_SUCCESS then
								idx = find(o[2], volumeMacros[1])
								if sum(activeChannels) = 0 then
									if idx <= 0 then
										volumeMacros[1] &= o[2]
										t = get_string()
										if equal(t, "=") then
											t = get_list()
											if t[1] = GET_SUCCESS then
												if in_range(t[2], -63, 63) and
												   in_range(t[3], -63, 63) then
													volumeMacros[2] = append(volumeMacros[2], t)
												else
													ERROR("Value out of range: " & sprint_list(t), lineNum)
												end if
											else
												ERROR("Bad volume macro: " & t, lineNum)
											end if
										else
											ERROR("Expected \'=\'", lineNum)
										end if
									else
										ERROR("Redefinition of @" & s, lineNum)
									end if
								else
									if idx <= 0 then
										ERROR("Undefined macro: @" & s, lineNum)
									else
										for i = 1 to length(supportedChannels) do
											if activeChannels[i] then
												if not is_empty_list(volumeMacros[2][idx]) then
													songs[songNum][i] &= {CMD_VOLMAC, idx}
												end if
											end if
										end for
									end if
								end if
							else
								ERROR("Syntax error: @" & s, lineNum)
							end if		


						elsif equal(substr(s, 1, 2), "te") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								write_note(1)
								if sum(activeChannels) = 0 then
									WARNING("Trying to set tone envelope with no active channels", lineNum)
								elsif integer(o[2]) then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsHwToneEnv[i] then
												if o[2] >= -supportsHwToneEnv[i] and o[2] <= supportsHwToneEnv[i] then
													songs[songNum][i] &= {CMD_HWTE, o[2]}
												else
													ERROR("Tone envelope number out of range: " & s, lineNum)
												end if
											else
												WARNING("Unsupported command for channel " & supportedChannels[i] & ": @" & s, lineNum)
											end if
										end if
									end for
								else
									ERROR("Bad tone envelope: " & s, lineNum)
								end if
							else
								ERROR("Bad tone envelope: " & s, lineNum)
							end if

						elsif equal(substr(s, 1, 2), "ve") then
							o = value(substr(s, 3, length(s)))
							if o[1] = GET_SUCCESS then
								write_note(1)
								if sum(activeChannels) = 0 then
									WARNING("Trying to set volume envelope with no active channels", lineNum)
								elsif integer(o[2]) then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsHwVolEnv[i] then
												if o[2] >= -supportsHwVolEnv[i] and o[2] <= supportsHwVolEnv[i] then
													songs[songNum][i] &= {CMD_HWVE, o[2]}
												else
													ERROR("Volume envelope number out of range: " & s, lineNum)
												end if
											else
												WARNING("Unsupported command for channel " & supportedChannels[i] & ": @" & s, lineNum)
											end if
										end if
									end for
								else
									ERROR("Bad volume envelope: " & s, lineNum)
								end if
							else
								ERROR("Bad volume envelope: " & s, lineNum)
							end if
							
						elsif equal(substr(s, 1, 1), "q") then
							o = value(substr(s, 2, length(s)))
							if o[1] = GET_SUCCESS then
								write_note(1)
								if sum(activeChannels) = 0 then
									WARNING("Trying to set cutoff with no active channels", lineNum)
								elsif integer(o[2]) and o[2] >= 0 and o[2] <= 15 then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											currentCutoff[i] = -o[2]
										end if
									end for
								else
									ERROR("Bad cutoff: " & s, lineNum)
								end if
							else
								ERROR("Bad cutoff: " & s, lineNum)
							end if
				
						else
							ERROR("Syntax error: @" & s, lineNum)
						end if
					end if
				else
					ERROR("Unexpected end of file", lineNum)
				end if

			-- Macro definition/invokation
			elsif n = '$' then
				write_note(1)
				s = get_string_in_range(ALPHANUM)
				skip_whitespace()
				m = getch()
				if length(s) then
					-- Macro definition
					if m = '{' then
						idx = find(s, macros[1])
						if idx <= 0 then
							if sum(activeChannels) = 0 then
								macro = {}
								while n != ';' do
									n = getch()
									if n = '%' then
										t = get_numeric_string()
										o = value(t)
										if o[1] = GET_SUCCESS then
											macro = append(macro, {1, o[2]})
										else
											ERROR("Syntax error: " & t, lineNum)
										end if
										n = getch()
										if n != '%' then
											ERROR("Missing %%", lineNum)
										end if
									elsif n = '}' or n = -1 then
										exit
									elsif n = 10 then
										lineNum += 1
									elsif n != 13 then
										macro = append(macro, {0, n})
									end if
								end while
							else
								ERROR("Macro definitions are not allowed while channels are active", lineNum)
							end if
							macros[1] = append(macros[1], s)
							macros[2] = append(macros[2], macro)
						else
							ERROR("Macro already defined: " & s, lineNum)
						end if

					-- Macro invokation
					elsif m = '(' then
						macro = {}
						while 1 do
							t = get_string_until(",)\t " & {13,10})
							macro = append(macro, t)
							skip_whitespace()
							n = getch()
							if n = -1 or n = ')' then
								exit
							end if
						end while
						u = ""
						idx = find(s, macros[1])
						if idx > 0 then
							for i = 1 to length(macros[2][idx]) do
								if macros[2][idx][i][1] = 0 then
									u &= macros[2][idx][i][2]
								else
									if macros[2][idx][i][2] > 0 and
									   macros[2][idx][i][2] <= length(macro) then
										u &= macro[macros[2][idx][i][2]]
									else
										ERROR(sprintf("Macro parameter enum out of range: %d", 
										      macros[2][idx][i][2]), lineNum)
									end if
								end if
							end for
							if fileDataPos <= length(fileData) then
								fileData = fileData[1..fileDataPos-1] & u & fileData[fileDataPos..length(fileData)]
							else
								fileData &= u
							end if

							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
								end if
							end for
							if numChannels = 0 then
								ERROR("Trying to invoke a macro with no channels active", lineNum)
							end if
						else
							ERROR("Undefined macro: " & s, lineNum)
						end if
					else
						ERROR("Syntax error: " & m, lineNum)
					end if
				else
					ERROR("Expected an identifier following $", lineNum)
				end if

			-- Callback
			elsif n = '!' then
				write_note(1)
				s = get_string_until("(\t " & {13,10})
				skip_whitespace()
				n = getch()
				if n = '(' then
					idx = find(s, callbacks)
					if idx <= 0 then
						callbacks = append(callbacks, s)
						idx = length(callbacks)
					end if

					t = get_string_until(")\t " & {13,10})
					o = value(t)
					if o[1] = GET_SUCCESS then
						if o[1] = 0 then
							-- Never (Off)
							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
									songs[songNum][i] &= CMD_CBOFF
								end if
							end for
							if numChannels = 0 then
								ERROR("Trying to set a callback with no channels active", lineNum)
							end if
						elsif o[1] = 1 then
							-- Once
							numChannels = 0
							for i = 1 to length(supportedChannels) do
								if activeChannels[i] then
									numChannels += 1
									songs[songNum][i] &= {CMD_CBONCE, idx - 1}
								end if
							end for
							if numChannels = 0 then
								ERROR("Trying to set a callback with no channels active", lineNum)
							end if
						else
							ERROR("Bad callback frequency: " & s, lineNum)
						end if
					elsif equal(t, "EVERY-NOTE") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVNT, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					elsif equal(t, "EVERY-VOL-CHANGE") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVVC, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					elsif equal(t, "EVERY-VOL-MIN") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVVM, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					elsif equal(t, "EVERY-OCT-CHANGE") then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								songs[songNum][i] &= {CMD_CBEVOC, idx - 1}
							end if
						end for
						if numChannels = 0 then
							ERROR("Trying to set a callback with no channels active", lineNum)
						end if

					else
						ERROR("Bad callback frequency: " & s, lineNum)
					end if
					n = getch()
					if n != ')' then
						ERROR("Expected ): " & n, lineNum)
					end if
				else
					ERROR("Expected (: " & n, lineNum)
				end if

			-- Multiline comment
			elsif n = '/' then
				write_note(1)
				m = getch()
				if m = '*' then
					while n != -1 do
						n = getch()
						if n = '*' then
							m = getch()
							if m = '/' then
								exit
							else
								ungetch()
							end if
						elsif n = 10 then
							lineNum += 1
						end if
					end while
				else
					ERROR("Syntax error: " & n, lineNum)
				end if

			elsif n = '[' then
				write_note(1)
				for i = 1 to length(supportedChannels) do
					if activeChannels[i] then
						loopStack[i] = push(loopStack[i], {length(songs[songNum][i]) + 2, songLen[songNum][i]})
						if length(loopStack[i]) > maxLoopDepth then
							ERROR(sprintf("Maximum nesting level for loops is %d", maxLoopDepth), lineNum)
						end if
						songs[songNum][i] &= {CMD_LOPCNT, 0}
					end if
				end for

			elsif n = ']' then
				write_note(1)
				for i = 1 to length(supportedChannels) do
					if activeChannels[i] then
						s = pop(loopStack[i])
						loopStack[i] = s[2]
						if s[1][1] != 0 then
							t = get_numeric_string()
							o = value(t)
							if o[1] = GET_SUCCESS then
								if o[2] > 0 then
									songs[songNum][i][s[1][1]] = o[2]	-- Loop count
									songs[songNum][i] &= {CMD_DJNZ, and_bits(s[1][1], #FF), floor(s[1][1] / #100)}
									songLen[songNum][i] += (songLen[songNum][i] - s[1][2]) * (o[2] - 1)
								else
									ERROR("Bad loop count: " & t, lineNum)
								end if
							else
								ERROR("Expected a loop count: " & t, lineNum)
							end if
						else
							ERROR("Use of ] with no matching [ on channel " & supportedChannels[i], lineNum)
						end if
					end if
				end for

			-- Octave up/down		
			elsif n = '>' or n = '<' then
				write_note(1)
				m = 1
				if n = '<' then
					m = -1
				end if
				if sum(activeChannels) = 0 then
					WARNING("Trying to change octave with no active channels: " & n, lineNum)
				else
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if currentOctave[i] + m >= minOctave and
							   currentOctave[i] + m <= maxOctave then
								currentOctave[i] += m
								pendingOctChange[i] = m
							else
								ERROR(sprintf("Octave out of range: " & n & " (%d)", currentOctave[i] + m), lineNum)
							end if
						end if
					end for
				end if

			elsif n = 'l' then
				write_note(1)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set length with no active channels", lineNum)
					elsif integer(o[2]) and find(o[2], SUPPORTED_LENGTHS) > 0 then
						numChannels = 0
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								numChannels += 1
								currentLength[i] = floor(power(2, find(o[2], SUPPORTED_LENGTHS) - 1))
							end if
						end for
						if numChannels = 0 then
							WARNING("Trying to set length with no active channels", lineNum)
						end if
					else
						ERROR("Bad length: " & s, lineNum)
					end if
				else
					ERROR("Bad length: " & s, lineNum)
				end if

			-- Set octave
			elsif n = 'o' then
				write_note(0)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set octave with no active channels", lineNum)
					elsif integer(o[2]) and o[2] >= minOctave and o[2] <= maxOctave then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								currentOctave[i] = o[2]
								songs[songNum][i] &= or_bits(CMD_OCTAVE, currentOctave[i] - minOctave)
							end if
						end for
					else
						ERROR("Bad octave: " & s, lineNum)
					end if
				else
					ERROR("Bad octave: " & s, lineNum)
				end if

			-- Set cutoff
			elsif n = 'q' then
				write_note(1)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set cutoff with no active channels", lineNum)
					elsif integer(o[2]) and o[2] >= 0 and o[2] <= 8 then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								currentCutoff[i] = o[2]
							end if
						end for
					else
						ERROR("Bad cutoff: " & s, lineNum)
					end if
				else
					ERROR("Bad cutoff: " & s, lineNum)
				end if

			-- Set tempo
			elsif n = 't' then
				write_note(1)
				s = get_numeric_string()
				o = value(s)
				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set tempo with no active channels", lineNum)
					elsif integer(o[2]) and o[2] >= 0 and o[2] < maxTempo then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								currentTempo[i] = o[2]
							end if
						end for
					else
						ERROR("Bad tempo: " & s, lineNum)
					end if
				else
					ERROR("Bad tempo: " & s, lineNum)
				end if

			-- Set volume
			elsif n = 'v' then
				write_note(1)
				s = get_numeric_string()
				if length(s) then
					o = value(s)
				else
					m = getch()
					if m = '+' then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								if currentVolume[i] < maxVolume[i] then
									o = {GET_SUCCESS, currentVolume[i] + 1}
								else
									WARNING("Volume too high on channel " & supportedChannels[i], lineNum)
								end if
							end if
						end for
					elsif m = '-' then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								if currentVolume[i] > minVolume then
									o = {GET_SUCCESS, currentVolume[i] -  1}
								else
									WARNING("Volume too low on channel " & supportedChannels[i], lineNum)
								end if
							end if
						end for
					else
						ERROR("Expected +, - or a number: " & m, lineNum)
					end if
				end if

				if o[1] = GET_SUCCESS then
					if sum(activeChannels) = 0 then
						WARNING("Trying to set volume with no active channels", lineNum)
					elsif integer(o[2]) and o[2] >= minVolume then
						for i = 1 to length(supportedChannels) do
							if activeChannels[i] then
								if o[2] <= maxVolume[i] then
									currentVolume[i] = o[2]
									if supportsVolumeChange[i] then
										l = floor((currentVolume[i] / maxVolume[i]) * machineVolLimit[i])
										if l <= #0F then
											songs[songNum][i] &= or_bits(CMD_VOL2, l)
										else
											songs[songNum][i] &= {CMD_VOLSET, l}
										end if
									else
										-- TODO: Handle this case (e.g. NES triangle channel)
									end if
								else
									ERROR("Bad volume: " & s, lineNum)
								end if
							end if
						end for
					else
						ERROR("Bad volume: " & s, lineNum)
					end if
				else
					ERROR("Bad volume: " & s, lineNum)
				end if

			-- Single line comment
			elsif n = ';' then
				write_note(1)
				while n != -1 do
					n = getch()
					if n = 10 then
						lineNum += 1
						exit
					end if
				end while

			elsif n = 'L' then
				write_note(1)
				if sum(activeChannels) = 0 then
					WARNING("Trying to set a loop point with no active channels", lineNum)
				else
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if loopPoint[i] = -1 then
								loopPoint[i] = length(songs[songNum][i])
								songLoopLen[songNum][i] = songLen[songNum][i]
							else
								ERROR("Loop point already defined for channel " &
								      supportedChannels[i], lineNum)
							end if
						end if
					end for
				end if

			-- Reads notes (e.g. c d+ a+4 g-1.. e1^8^32 f&f16&f32)
			elsif find(n, NOTES) > 0 then
				if not slur then
					write_note(0)
				end if

				dot 	= 0
				tie 	= 0
				slur 	= 0
				hasTie 	= 0
				hasSlur = 0
				firstNote = -1
				len = -1
				
				l = 0
				while n != -1 do
					if find(n, NOTES) > 0 then
						if l > 0 and not slur then
							ungetch()
							exit
						end if
						note = find(n, NOTES)
					elsif n = '&' then
						if hasTie then
							ERROR("Trying to use & and ^ in same expression", lineNum)
						end if
						hasSlur = 1
						slur = 1
						note = -1
					elsif n = '^' then
						if hasSlur then
							ERROR("Trying to use & and ^ in same expression", lineNum)
						end if
						hasTie = 1
						tie = 1
					elsif n = '.' then
						dot = 1
					else
						ungetch()
						exit
					end if

					if (find(n, NOTES) > 0 and n != 'r') or (l > 0 and note != -1 and slur) then
						flatSharp = 0
						len = -1
						m = getch()
						--t = ""
						--while m = '+' or m = '-' do
						--	if m = '+' then
						--		flatSharp += 1
						--		t &= m
						--	elsif m = '-' then
						--		flatSharp -= 1
						--		t &= m
						--	else
						--		exit
						--	end if
						--	m = getch()
						--end while
						--ungetch()
						--if NOTES[remainder(note + flatSharp, 12) + 1] != -2
						--	ERROR("Bad note: " & n & t, lineNum)
						--end if
						if m = '+' then
							flatSharp = 1
							if NOTES[note + flatSharp] != -2 then
								ERROR("Bad note: " & n & "+", lineNum)
							end if
						elsif m = '-' then
							flatSharp = -1
							if note + flatSharp < 1 or NOTES[note + flatSharp] != -2 then
								ERROR("Bad note: " & n & "-", lineNum)
							end if
						else
							ungetch()
						end if
					elsif n = 'r' then
						flatSharp = 0
					end if

					if firstNote = -1 then
						firstNote = note + flatSharp
					else
						if firstNote != note + flatSharp and note != -1 then
							ERROR("Trying to concatenate different notes", lineNum)
						end if
					end if

					if find(n, NOTES) > 0 or tie then
						s = get_numeric_string()
						if length(s) then
							o = value(s)
							if o[1] = GET_SUCCESS then
								if find(o[2], SUPPORTED_LENGTHS) > 0 then
									len = floor(power(2, find(o[2], SUPPORTED_LENGTHS) - 1))
								else
									ERROR("Unsupported length: " & s, lineNum)
								end if
							else
								ERROR("Bad length: " & s, lineNum)
							end if
						end if
					end if
					if len = -1 and tie then
						ERROR("Expected a length", lineNum)
					end if

					numChannels = 0
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if currentOctave[i] <= minOctave and
							   note < minNote then
								WARNING("The frequency of the note can not be reproduced for this octave: " & NOTES[note], lineNum)
							end if
							numChannels += 1
							if len = -1 then
								len = currentLength[i]
							end if
							if l = 0 then
								if n != 'r' then
									currentNote[i] = {(currentOctave[i] - minOctave) * 12 + note + flatSharp - 1, len}
								else
									currentNote[i] = {rest, len}
								end if
								lastSetLength[i] = len
							else
								if dot then
									len = floor(lastSetLength[i] / 2)
									lastSetLength[i] = len
									if len >= 1 then
										currentNote[i][2] += len
									else
										ERROR("Note length out of range due to dot command", lineNum)
									end if
									dot = 0
								elsif tie then
									currentNote[i][2] += len
									tie = 0
								elsif slur then
									if note != -1 then
										if (currentOctave[i] - minOctave) * 12 + note + flatSharp - 1 = currentNote[i][1] then
											currentNote[i][2] += len
											lastSetLength[i] = len
											slur = 0
										else
											ERROR("Bad note: " & n, lineNum)
										end if
									end if
								else
									ERROR("Bad note: " & n, lineNum)
								end if
							end if
						end if
					end for
					if numChannels = 0 then
						ERROR("Trying to play a note with no active channels", lineNum)
					end if

					n = getch()
					l += 1
				end while
				tie = 0
				slur = 0
				dot = 0
				lastSetLength = repeat(0, length(supportedChannels))

			elsif n = '{' then
				write_note(1)
				if sum(activeChannels) = 0 then
					ERROR("Trying to play a tuple with no active channels", lineNum)
				else
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if integer(tuple[i]) then
								tuple[i] = {}
							else
								ERROR("Nested tuples are not allowed", lineNum)
							end if
						end if
					end for
				end if
				
			elsif n = '}' then
				write_note(1)
				s = get_numeric_string()
				len = -1
				if length(s) then
					o = value(s)
					if o[1] = GET_SUCCESS then
						if find(o[2], SUPPORTED_LENGTHS) > 0 then
							len = floor(power(2, find(o[2], SUPPORTED_LENGTHS) - 1))
						else
							ERROR("Unsupported length: " & s, lineNum)
						end if
					else
						ERROR("Bad length: " & s, lineNum)
					end if
				end if
				if len = -1 then
					ERROR("No length specified for tuple", lineNum)
				end if
				if sum(activeChannels) = 0 then
					ERROR("Trying to close a tuple with no active channels", lineNum)
				else
					numChannels = 0
					for i = 1 to length(supportedChannels) do
						if activeChannels[i] then
							if sequence(tuple[i]) then
								numChannels += 1
								write_tuple(i, tuple[i], len)
								tuple[i] = 0
							end if
						end if
					end for
					if numChannels = 0 then
						ERROR("Found } with no matching {", lineNum)
					end if
				end if
							
				
			elsif find(n, supportedChannels) > 0 or find(n, "CDEMPW") > 0 then
				write_note(1)
				if n = 'C' then
					m = getch()
					if m = 'S' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], panMacros[1])
							if sum(activeChannels) = 0 then
								ERROR("CS requires at least one active channel", lineNum)
							elsif idx > 0 then
								if supportsPan then
									if sum(activeChannels) > 0 then
										for i = 1 to length(supportedChannels) do
											if activeChannels[i] then
												songs[songNum][i] &= {CMD_PANMAC, idx}
											end if
										end for
									else
										WARNING("Trying to use CS with no channels active", lineNum)
									end if
								else
									WARNING("Unsupported command for this target: CS", lineNum)
								end if
							elsif supportsPan then
								ERROR("Undefined macro: CS" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								if supportsPan then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											songs[songNum][i] &= {CMD_PANMAC, 0}
										end if
									end for
								end if
							else
								ERROR("Syntax error: CS" & s, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if						
				
				elsif n = 'D' then
					m = getch()
					if m = '-' or is_numeric(m) then
						bypass = 1
						ungetch()
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							if sum(activeChannels) = 0 then
								ERROR("Detune requires at least one active channel", lineNum)
							else
								if in_range(o[2], -127, 127) then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsDetune[i] then
												songs[songNum][i] &= {CMD_DETUNE, o[2]}
											else
												WARNING("Unsupported command for this channel: D", lineNum)
											end if
										end if
									end for
								else
									ERROR("Detune value out of range: " & s, lineNum)
								end if
							end if
						else
							ERROR("Expected a number: " & s, lineNum)
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if						
						
				elsif n = 'E' then
					m = getch()
					if m = 'N' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], arpeggios[1])
							if sum(activeChannels) = 0 then
								ERROR("EN requires at least one active channel", lineNum)
							elsif idx > 0 then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if not is_empty_list(arpeggios[2][idx]) then
											if enRev = 0 then
												songs[songNum][i] &= {CMD_ARPMAC, idx}
												usesEN[1] += 1
											else
												songs[songNum][i] &= {CMD_APMAC2, idx}
												usesEN[2] += 1
											end if
										end if
									end if
								end for
							else
								ERROR("Undefined macro: EN" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if enRev = 0 then
											songs[songNum][i] &= CMD_ARPOFF 
										else
											songs[songNum][i] &= CMD_ARPOFF 
										end if
									end if
								end for
							else
								ERROR("Syntax error: EN" & s, lineNum)
							end if
						end if

					elsif m = 'P' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], pitchMacros[1])
							if sum(activeChannels) = 0 then
								ERROR("EP requires at least one active channel", lineNum)
							elsif idx > 0 then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if not is_empty_list(pitchMacros[2][idx]) then
											songs[songNum][i] &= {CMD_SWPMAC, idx}
										end if
									end if
								end for
							else
								ERROR("Undefined macro: EP" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then	
										songs[songNum][i] &= {CMD_SWPMAC, 0}
									end if
								end for
							else
								ERROR("Syntax error: EP" & s, lineNum)
							end if
						end if

					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				elsif n = 'M' then
					m = getch()
					if m = 'P' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], vibratos[1])
							if sum(activeChannels) = 0 then
								ERROR("MP requires at least one active channel", lineNum)
							elsif idx > 0 then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										songs[songNum][i] &= {CMD_VIBMAC, idx}
									end if
								end for
							else
								ERROR("Undefined macro: MP" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then	
										songs[songNum][i] &= {CMD_VIBMAC, 0}
									end if
								end for
							else
								ERROR("Syntax error: MP" & s, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				elsif n = 'P' then
					m = getch()
					if m = 'T' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], portamentos[1])
							if sum(activeChannels) = 0 then
								ERROR("PT requires at least one active channel", lineNum)
							elsif idx > 0 then
								-- TODO: set portamento
								WARNING("PT command not yet implemented", lineNum)
							else
								ERROR("Undefined macro: PT" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								-- TODO: deactivate portamento
							else
								ERROR("Syntax error: PT" & s, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if

				elsif n = 'W' then
					m = getch()
					if m = 'T' then
						bypass = 1
						s = get_numeric_string()
						o = value(s)
						if o[1] = GET_SUCCESS then
							idx = find(o[2], waveforms[1])
							if sum(activeChannels) = 0 then
								ERROR("WT requires at least one active channel", lineNum)
							elsif idx > 0 then
								if sum(activeChannels) > 0 then
									for i = 1 to length(supportedChannels) do
										if activeChannels[i] then
											if supportsWave[i] then
												songs[songNum][i] &= {CMD_LDWAVE, idx}
											else
												WARNING("Unsupported command for channel " & supportedChannels[i] & ": WT", lineNum)
											end if
										end if
									end for
								else
									WARNING("Trying to use WT with no channels active", lineNum)
								end if
							elsif sum(supportsWave) != 0 then
								ERROR("Undefined macro: WT" & s, lineNum)
							end if
						else
							m = getch()
							t = m & getch()
							if equal(t, "OF") then
								for i = 1 to length(supportedChannels) do
									if activeChannels[i] then
										if supportsWave[i] then
											songs[songNum][i] &= {CMD_LDWAVE, 0}
										else
											WARNING("WT commands not supported for channel " & supportedChannels[i], lineNum)
										end if
									end if
								end for
							else
								ERROR("Syntax error: WT" & s, lineNum)
							end if
						end if
					else
						ungetch()
						if find(n, supportedChannels) <= 0 then
							ERROR("Unexpected character: " & n, lineNum)
						end if
					end if
					
				end if
				if not bypass then
					if not lastWasChannelSelect then
						activeChannels = repeat(0, length(supportedChannels))
					end if
					activeChannels[find(n, supportedChannels)] = 1
				end if		
			elsif find(n, "\t " & {13, 10}) > 0 then
				write_note(n = 10)	
			else
				if n = '%' then
					ERROR("Unexpectd character: %%", lineNum)
				else
					ERROR("Unexpected character: " & n, lineNum)
				end if
			end if

			if find(n2, supportedChannels) > 0 then
				lastWasChannelSelect = 1
			else
				lastWasChannelSelect = 0
			end if
		end if
	end while	
	
	write_note(1)
	
	for i = 1 to length(loopStack) do
		if length(loopStack[i]) then
			ERROR("Open [ on channel " & supportedChannels[i], lineNum)
		end if
	end for

	for i = 1 to length(supportedChannels) do
		songLoopLen[songNum][i] = songLen[songNum][i] - songLoopLen[songNum][i]
		if loopPoint[i] = -1 then
			songs[songNum][i] &= CMD_END
		else
			if not hasAnyNote[i] then
				songs[songNum][i] &= CMD_END
			else
				songs[songNum][i] &= {CMD_JMP, and_bits(loopPoint[i], #FF), floor(loopPoint[i] / #100)}
			end if
		end if
	end for
	
	m = -1
	for j = 1 to length(songs) do
		if sequence(songs[j]) then
			for i = 1 to length(supportedChannels) do
				if songLen[j][i] then
					if m = -1 then
						m = songLen[j][i]
					elsif songLen[j][i] != m then
						WARNING(sprintf("Mismatch in length between channels in song %d", j), -1)
						exit
					end if
				end if
			end for
		end if
	end for
	
	if target = TARGET_SMS or target = TARGET_SGG then
		output_sms(writeVGM)
	elsif target = TARGET_GBC then
		output_gbc()
	elsif target = TARGET_NDS then
		output_nds()
	end if
	
	puts(1, "\nDone.\n")
end procedure



verbose = 0
warningsAreErrors = 0
target = TARGET_UNKNOWN

procedure show_help(sequence what)
	if length(what) = 0 then
		puts(1, "Usage: xpmc.exe [options] <target> input [output]\n\n")
		puts(1, "Options:\n")
		puts(1, "\t-h\tShow this information\n") -- / Information about a\n\t                specific command (e.g. -h EP)\n")
		puts(1, "\t-v\tVerbose mode\n")
		puts(1, "\t-w\tTreat warnings as errors\n")
		puts(1,"\nTarget:\n")
		--puts(1,"\t-gba\tGameboy Advance\n")
		--puts(1,"\t-gbc\tGameboy / Gameboy Color\n")
		--puts(1,"\t-gen\tSEGA Genesis\n")
		--puts(1,"\t-nds\tNintendo DS\n")
		--puts(1,"\t-sat\tSEGA Saturn\n")
		puts(1,"\t-sgg\tSEGA Game Gear\n")
		puts(1,"\t-sms\tSEGA Master System\n")
	elsif equal(what, "EN") then
		puts(1,"EN:\tArpeggio macro: { num | num}. num is zero or more numbers in\n\tthe range -63-63.\n")
	end if
end procedure


cmdLine = command_line()
if length(cmdLine) <= 2 then
	show_help("")
	abort(0)
end if
for i=3 to length(cmdLine) do
	if cmdLine[i][1] = '-' then
		restOfLine = cmdLine[i][2..length(cmdLine[i])]
		if equal(restOfLine, "v") then
			verbose = 1
		elsif equal(restOfLine, "w") then
			warningsAreErrors = 1
		elsif equal(restOfLine, "h") or equal(restOfLine, "help") then
			if i < length(cmdLine) then
				if equal(cmdLine[i+1], "EN") then
					show_help("EN")
				elsif equal(cmdLine[i+1], "EP") then
					show_help("EP")
				elsif equal(cmdLine[i+1], "MP") then
					show_help("MP")
				elsif equal(cmdLine[i+1], "VS") then
					show_help("VS")
				else
					show_help("")
				end if
			else
				show_help("")
			end if
			exit
		elsif equal(restOfLine, "sms") then
			target = TARGET_SMS
		elsif equal(restOfLine, "sgg") then
			target = TARGET_SGG
		--elsif equal(restOfLine, "gbc") then
		--	target = TARGET_GBC
		--elsif equal(restOfLine, "gba") then
		--	target = TARGET_GBA
		--elsif equal(restOfLine, "nds") then
		--	target = TARGET_NDS
		--elsif equal(restOfLine, "sat") then
		--	target = TARGET_SAT
		--elsif equal(restOfLine, "xgs") then
		--	target = TARGET_XGS
		end if
	else
		compile(cmdLine[i..length(cmdLine)])
		exit
	end if
end for
